from ..conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ....utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval
from ..agent import Agent
import ray
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from .. import get_agent_name,run_agent_func,ChatResponse
from byzerllm.apps.agent.extensions.simple_retrieval_client import SimpleRetrievalClient
import uuid
import json
from byzerllm.apps.llama_index import get_service_context,get_storage_context
from llama_index import VectorStoreIndex
from llama_index.query_engine import SubQuestionQueryEngine


try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x

from llama_index.tools import QueryEngineTool, ToolMetadata    



class LlamaIndexSubQuestionAgent(ConversableAgent): 
    PROMPT_DEFAULT = """You're a retrieve augmented chatbot. """    
    DEFAULT_SYSTEM_MESSAGE = PROMPT_DEFAULT
    
    def __init__(
        self,
        name: str,
        llm: ByzerLLM,        
        retrieval: ByzerRetrieval,        
        chat_name:str,
        owner:str,                
        update_context_retry: int = 3,        
        system_message: Optional[str] = DEFAULT_SYSTEM_MESSAGE,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):       
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )
        self.chat_name = chat_name
        self.owner = owner        
        self.update_context_retry = update_context_retry


        self._reply_func_list = []
        # self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)   
        self.register_reply([Agent, ClientActorHandle,str], LlamaIndexSubQuestionAgent.generate_retrieval_based_reply) 
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply) 
        self.service_context = get_service_context(llm)
        self.storage_context = get_storage_context(llm,retrieval)        
              
        
                        

    def generate_retrieval_based_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:  
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
                
        new_message = messages[-1]
        index = VectorStoreIndex.from_vector_store(vector_store = self.storage_context.vector_store,service_context=self.service_context)
        vector_query_engine = index.as_query_engine()
        query_engine_tools = [
                                QueryEngineTool(
                                    query_engine=vector_query_engine,
                                    metadata=ToolMetadata(
                                        name="common",
                                        description="common",
                                    ),
                                                ),
                            ]                

        query_engine = SubQuestionQueryEngine.from_defaults(
            query_engine_tools=query_engine_tools,
            service_context=self.service_context,
            use_async=True,
        )
        response = query_engine.query(new_message["content"])        
        return True, {
            "content":response.response,
            "metadata":{"agent":self.name,"TERMINATE":True}
        }
        
                
        
                    


from langchain.embeddings.base import Embeddings
from typing import List, Union

try:
    import torch
    import torch.nn.functional as F
    from transformers import pipeline

    class ByzerSentenceTransformerEmbeddings(Embeddings):

        def __init__(self, model, tokenizer, device="auto"):
            if model:
                self.model = model
            if tokenizer:
                self.model = tokenizer

            if device == "auto":
                self.device = (
                    torch.cuda.current_device() if torch.cuda.is_available() else "cpu"
                )
            else:
                self.device = device

        def _encode(self, texts: List[str], extract_params={}):
            embeddings = [
                emb.tolist() for emb in self.model.encode(texts, **extract_params)
            ]
            return embeddings

        def embed_documents(
            self, texts: List[str], extract_params={}
        ) -> List[List[float]]:
            embeddings = self._encode(texts, extract_params)
            return embeddings

        def embed_query(self, text: str, extract_params={}) -> List[float]:
            embedding = self._encode([text], extract_params)
            return embedding[0]

    class ByzerLLMEmbeddings(Embeddings):
        def __init__(
            self, model, tokenizer, device="auto", use_feature_extraction=False
        ):
            self.model = model
            self.tokenizer = tokenizer

            if device == "auto":
                self.device = (
                    torch.cuda.current_device() if torch.cuda.is_available() else "cpu"
                )
            else:
                self.device = device

            self.pipeline = None
            if use_feature_extraction:
                self.pipeline = pipeline(
                    "feature-extraction", model=model, tokenizer=tokenizer, device=0
                )

        def _encode(self, texts: List[str], extract_params={}):
            if self.pipeline:
                return [self.pipeline(text)[0][-1] for text in texts]
            else:
                _, embeddings = self.get_embedding_with_token_count(texts)
                embeddings = embeddings.detach().cpu()
                embeddings = embeddings.numpy()
                embeddings = [emb.tolist() for emb in embeddings]
                return embeddings

        def embed_documents(
            self, texts: List[str], extract_params={}
        ) -> List[List[float]]:
            embeddings = self._encode(texts, extract_params)
            return embeddings

        def embed_query(self, text: str, extract_params={}) -> List[float]:
            embedding = self._encode([text], extract_params)
            return embedding[0]

        # copied from https://huggingface.co/sentence-transformers/all-MiniLM-L6-v2#usage-huggingface-transformers
        def get_embedding_with_token_count(
            self,
            sentences: Union[str, List[str]],
            ignore_keys: List[str] = ["token_type_ids"],
        ):
            # Mean Pooling - Take attention mask into account for correct averaging
            def mean_pooling(model_output, attention_mask):
                # First element of model_output contains all token embeddings
                token_embeddings = model_output[0]
                input_mask_expanded = (
                    attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
                )
                return torch.sum(
                    token_embeddings * input_mask_expanded, 1
                ) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)

            # Tokenize sentences
            encoded_input = self.tokenizer(
                sentences, padding=True, truncation=True, return_tensors="pt"
            )
            inputs = encoded_input.to(self.device)
            token_count = inputs["attention_mask"].sum(dim=1).tolist()[0]
            # Compute token embeddings

            for ignore_key in ignore_keys:
                if hasattr(inputs, ignore_key):
                    del inputs[ignore_key]

            model_output = self.model(**inputs)
            # Perform pooling
            sentence_embeddings = mean_pooling(model_output, inputs["attention_mask"])
            # Normalize embeddings
            sentence_embeddings = F.normalize(sentence_embeddings, p=2, dim=1)

            return token_count, sentence_embeddings

except ImportError:

    class ByzerLLMEmbeddings:
        def __init__(self, *args, **kwargs):
            pass

        def embed_documents(self, *args, **kwargs):
            raise ImportError("transformers is not installed")

        def embed_query(self, *args, **kwargs):
            raise ImportError("transformers is not installed")

    class ByzerSentenceTransformerEmbeddings:
        def __init__(self, *args, **kwargs):
            pass

        def embed_documents(self, *args, **kwargs):
            raise ImportError("transformers is not installed")

        def embed_query(self, *args, **kwargs):
            raise ImportError("transformers is not installed")



from typing import List,Dict,Annotated,Any
from byzerllm.utils.client import ByzerLLM,message_utils,code_utils,LLMRequest
from byzerllm.utils.retrieval import ByzerRetrieval
import json
import numpy as np
import copy
import pydantic
from . import QueryRewriteResult,Action

class QueryContext:
    '''
    this tool is used to extract the key messages from the user's question which
    is used to generate the sql code conditions e.g. filter conditions, group by conditions, order by conditions
    or aggregate fields.
    '''
    def __init__(self,llm:ByzerLLM, retrieval:ByzerRetrieval,
                 sys_message:List[Dict[str,Any]],
                 messages:List[Dict[str,Any]],**kwargs):
        self.messages = messages
        self._system_message = sys_message
        self.llm = llm
        self.retrieval = retrieval
        self.params = kwargs

    def apply(self):        
        m = copy.deepcopy(self.messages[-1])
        temp_conversation = [{"role":"user","content":'''
首先，根据我们前面几条聊天内容，针对我现在的问题，进行一个信息扩充。如果我的问题信息已经比较充足，
则输出原有问题即可。
     
注意：
1. 不要询问用户问题          
2. 不要生成SQL
3. 不要额外添加上下文中不存在的信息
4. 不要关注时间,不要改写时间
6. 尽量保证信息完整 
'''}]
        class SingleLine(pydantic.BaseModel):
            content:str=pydantic.Field(...,description="改写后的query")

        t = self.llm.chat_oai(
            conversations=message_utils.padding_messages_merge(self._system_message + self.messages + self.params.get("temp_conversation",temp_conversation)),
            response_class=SingleLine,
            enable_default_sys_message=True
            ) 
        new_query = m["content"]           
        if t[0].value:
            new_query = t[0].value.content               
            print(f'context query rewrite: {m["content"]} -> {new_query}\n\n',flush=True)          
            m["content"] = new_query   
    #         if new_query != m["content"]:
    #             temp1 = self.llm.emb(None,LLMRequest(instruction=new_query))
    #             temp2 = self.llm.emb(None,LLMRequest(instruction=m["content"]))
    #             sim = np.dot(temp1[0].output,temp2[0].output)
    #             if sim > 0.8:
    #                 print(f'context query rewrite: {m["content"]} -> {new_query}\n\n',flush=True)
    #                 m["content"] = new_query                    
    #             else:
    #                 print(f'''context query rewrite fail. 
    # the similarity is too low {sim}
    # query:  {m["content"]}
    # new_query: {new_query}
    # \n\n''',flush=True)

        return QueryRewriteResult(message = m,action = Action.CONTINUE,extra_info={"new_query":new_query})       
             


    


from typing import List,Dict
import json
from transformers import AutoTokenizer, BitsAndBytesConfig
from byzerllm.utils.metrics import Metric
from peft import LoraConfig, get_peft_model, prepare_model_for_kbit_training
from transformers import (
    set_seed,
    HfArgumentParser,
    TrainingArguments,
    AutoModelForCausalLM,
    AutoModel
)
import argparse
import os
from os.path import join
import torch
import bitsandbytes as bnb
from collections import defaultdict

from .collator import SFTDataCollator
from .dataset import SFTDataset
from .argument import QLoRAArguments
from .trainer import LoRATrainer
from .loss import TargetLMLoss


def verify_model_dtype(model):
    """
    查看模型种各种类型的参数的情况
    """
    dtype2param_num = defaultdict(int)  # 每种数据类型的参数量
    dtype2param_name = defaultdict(list)  # 每种数据类型的参数名称
    dtype2trainable_param_num = defaultdict(int)  # 每种数据类型参与训练的参数量
    dtype2trainable_param_name = defaultdict(list)  # 每种数据类型参与训练的参数名称
    for name, p in model.named_parameters():
        dtype = p.dtype
        dtype2param_num[dtype] += p.numel()
        dtype2param_name[dtype].append(name)
        if p.requires_grad:
            dtype2trainable_param_num[dtype] += p.numel()
            dtype2trainable_param_name[dtype].append(name)
    # 统计全部参数中，各种类型参数分布
    total = 0
    print('verify all params of the model')
    for k, v in dtype2param_num.items():
        total += v
    for k, v in dtype2param_num.items():
        print(k, v, v / total)
    for k, v in dtype2trainable_param_name.items():
        print(k, v)

    print()
    # 统计可训练参数中，各种类型参数分布
    print('verify trainable params the model')
    total_trainable = 0
    for k, v in dtype2trainable_param_num.items():
        total_trainable += v
    for k, v in dtype2trainable_param_num.items():
        print(k, v, v / total_trainable)
    for k, v in dtype2trainable_param_num.items():
        print(k, v)


def find_all_linear_names(model):
    """
    找出所有全连接层，为所有全连接添加adapter
    """
    cls = bnb.nn.Linear4bit
    lora_module_names = set()
    for name, module in model.named_modules():
        if isinstance(module, cls):
            names = name.split('.')
            lora_module_names.add(names[0] if len(names) == 1 else names[-1])

    if 'lm_head' in lora_module_names:  # needed for 16-bit
        lora_module_names.remove('lm_head')
    return list(lora_module_names)


def setup_everything(lora_config:str, args:List[str]):
    parser = argparse.ArgumentParser()    
    args = parser.parse_args(args=args)
    
    # 读取训练的参数配置
    parser = HfArgumentParser((QLoRAArguments, TrainingArguments))
    # 解析得到自定义参数，以及自带参数
    args, training_args = parser.parse_dict(json.loads(lora_config),allow_extra_keys=True)
    # 设置随机种子
    set_seed(training_args.seed)
    return args, training_args


def init_components(args, training_args,extra_params):
    """
    初始化各个组件
    """
    print('Initializing components...')
    # 下面的设置至关重要，否则无法多卡训练
    world_size = int(os.environ.get("WORLD_SIZE", 1))
    ddp = world_size != 1
    # training_args.ddp_find_unused_parameters = False if ddp else None
    device_map = "auto"
    # if we are in a distributed setting, we need to set the device map and max memory per device
    if os.environ.get('LOCAL_RANK') is not None:
        local_rank = int(os.environ.get('LOCAL_RANK', '0'))
        device_map = {'': local_rank}
    # 加载tokenzier
    tokenizer = AutoTokenizer.from_pretrained(
        args.model_name_or_path,
        trust_remote_code=True,
    )
    # 部分tokenizer没有pad_token_id
    if tokenizer.pad_token_id is None:
        if tokenizer.unk_token_id is None:
            print(f"tokenizer has no pad_token_id and unk_token_id, setting it to 0")
            tokenizer.pad_token_id = 0
        else:                
            print(f"tokenizer has no pad_token_id({tokenizer.pad_token_id}), setting it to unk_token_id({tokenizer.unk_token_id})")
            tokenizer.pad_token_id = tokenizer.unk_token_id
    # 如果两者相同，模型训练时不会计算eos_token_id的loss
    if tokenizer.pad_token_id == tokenizer.eos_token_id:
        raise Exception('pad_token_id should not be equal to eos_token_id')       
    
    # 加载模型  
    model_type = extra_params.get("model_type","casual_lm")

    if model_type == "casual_lm":
        model = AutoModelForCausalLM.from_pretrained(
            args.model_name_or_path,
            device_map=device_map,
            load_in_4bit=True,
            torch_dtype=torch.float16,
            trust_remote_code=True,
            quantization_config=BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_compute_dtype=torch.float16,
                bnb_4bit_use_double_quant=True,
                bnb_4bit_quant_type="nf4",
                llm_int8_threshold=6.0,
                llm_int8_has_fp16_weight=False,
            ),
        )    
        model.tie_weights()
    else:
        model = AutoModel.from_pretrained(
            args.model_name_or_path,
            device_map=device_map,
            load_in_4bit=True,
            torch_dtype=torch.float16,
            trust_remote_code=True,
            quantization_config=BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_compute_dtype=torch.float16,
                bnb_4bit_use_double_quant=True,
                bnb_4bit_quant_type="nf4",
                llm_int8_threshold=6.0,
                llm_int8_has_fp16_weight=False,
            ),
        )     
    # casts all the non int8 modules to full precision (fp32) for stability
    model = prepare_model_for_kbit_training(model, use_gradient_checkpointing=training_args.gradient_checkpointing)    
    print(f'memory footprint of model: {model.get_memory_footprint()/(1024*1024*1024)} GB')
    # 找到所有需要插入adapter的全连接层
    target_modules = find_all_linear_names(model)
    # 初始化lora配置
    config = LoraConfig(
        r=args.lora_rank,
        lora_alpha=args.lora_alpha,
        target_modules=target_modules,
        lora_dropout=args.lora_dropout,
        bias="none",
        task_type="CAUSAL_LM",
    )
    model = get_peft_model(model, config)
    model.print_trainable_parameters()
    model.config.torch_dtype = torch.float32

    # 查看模型种各种类型的参数的情况
    verify_model_dtype(model)

    # 初始化损失函数
    print(f'Initializing loss function (ignore_index={tokenizer.pad_token_id})...',flush=True)
    loss_func = TargetLMLoss(ignore_index=tokenizer.pad_token_id)
    
    train_dataset = SFTDataset(args.train_file, tokenizer, args.max_seq_length, **extra_params)
    data_collator = SFTDataCollator(tokenizer, args.max_seq_length, **extra_params)

    # 初始化Trainer
    trainer = LoRATrainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset,
        tokenizer=tokenizer,
        data_collator=data_collator,
        compute_loss=loss_func
    )
    return trainer


def train(lora_config:str, args:List[str],extra_params={})->str:

    sft_name = extra_params["sft_name"]   
    # 进行一些配置和检查
    parsed_args, training_args = setup_everything(lora_config, args)
    # 加载各种组件
    trainer = init_components(parsed_args, training_args,extra_params)
    # 开始训练
    print(f"*** [{sft_name}] starting training ***")
    train_result = trainer.train()

    # 打印 token 总数
    print(f"[{sft_name}] total tokens: {trainer.train_dataset.dataset_tokens_count}",flush=True)
    token_metrics = Metric()
    token_metrics.inc(f"sft_{sft_name}_tokens_num",trainer.train_dataset.dataset_tokens_count)
    token_metrics.push()

    # 保存最好的checkpoint
    final_save_path = join(training_args.output_dir, 'final')
    trainer.save_model(final_save_path)  # Saves the tokenizer too
    # 保存训练指标
    metrics = train_result.metrics
    trainer.log_metrics("train", metrics)
    trainer.save_metrics("train", metrics)
    trainer.save_state()  
    return final_save_path  





import json
from typing import Dict, List, Union, Any
from tokenizers import Tokenizer


def get_meta(self):
    return [
        {
            "model_deploy_type": "proprietary",
            "backend": "tokenizers",
            "message_format": True,
        }
    ]


def process_input(ins: Union[str, List[Dict[str, Any]], Dict[str, Any]]):

    if isinstance(ins, list) or isinstance(ins, dict):
        return ins

    content = []
    try:
        ins_json = json.loads(ins)
    except:
        return ins

    ## 如果是字典，应该是非chat格式需求，比如语音转文字等
    if isinstance(ins_json, dict):
        return ins_json

    if isinstance(ins_json, list):
        if ins_json and isinstance(ins_json[0], dict):
            # 根据key值判断是什么类型的输入，比如语音转文字，语音合成等
            for temp in ["input", "voice", "audio", "audio_url"]:
                if temp in ins_json[0]:
                    return ins_json[0]

    content = []
    #     [
    #     {"type": "text", "text": "What’s in this image?"},
    #     {
    #       "type": "image_url",
    #       "image_url": {
    #         "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
    #         "detail": "high"
    #       },
    #     },
    #   ],
    for item in ins_json:
        # for format like this: {"image": "xxxxxx", "text": "What’s in this image?","detail":"high"}
        # or {"image": "xxxxxx"}, {"text": "What’s in this image?"}
        if ("image" in item or "image_url" in item) and "type" not in item:
            image_data = item.get("image", item.get("image_url", ""))
            ## "data:image/jpeg;base64,"
            if not image_data.startswith("data:"):
                image_data = "data:image/jpeg;base64," + image_data

            ## get the other fields except image/text/image_url
            other_fields = {
                k: v for k, v in item.items() if k not in ["image", "text", "image_url"]
            }
            content.append(
                {
                    "image_url": {"url": image_data, **other_fields},
                    "type": "image_url",
                }
            )

        if "text" in item and "type" not in item:
            text_data = item["text"]
            content.append({"text": text_data, "type": "text"})

        ## for format like this: {"type": "text", "text": "What’s in this image?"},
        ## {"type": "image_url", "image_url": {"url":"","detail":"high"}}
        ## this is the standard format, just return it
        if "type" in item and item["type"] == "text":
            content.append(item)

        if "type" in item and item["type"] == "image_url":
            content.append(item)

    if not content:
        return ins

    return content


def stream_chat(
    self,
    tokenizer,
    ins: str,
    his: List[Dict[str, str]] = [],
    max_length: int = 4090,
    top_p: float = 0.95,
    temperature: float = 0.1,
    **kwargs,
):
    messages = [
        {"role": message["role"], "content": process_input(message["content"])}
        for message in his
    ] + [{"role": "user", "content": process_input(ins)}]

    encoded = self.encode(json.dumps(messages, ensure_ascii=False))
    return [
        (
            f"{len(encoded.tokens)}",
            {
                "metadata": {
                    "request_id": "",
                    "input_tokens_count": len(encoded.tokens),
                    "generated_tokens_count": 0,
                    "time_cost": 0,
                    "first_token_time": 0,
                    "speed": 0,
                }
            },
        )
    ]


def init_model(
    model_dir, infer_params: Dict[str, str] = {}, sys_conf: Dict[str, str] = {}
):

    model = Tokenizer.from_file(model_dir)
    import types

    model.stream_chat = types.MethodType(stream_chat, model)
    model.get_meta = types.MethodType(get_meta, model)

    return (model, None)


# Helper functions from web_ui.py



from pyjava import PythonContext, RayContext
from typing import Dict, Any, List, Optional, Union, Tuple, Callable, Annotated
from pyjava.udf import UDFBuilder
import ray
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from byzerllm.utils.client import code_utils
from byzerllm.utils import (
    function_calling_format,
    response_class_format,
    response_class_format_after_chat,
    FunctionCallList,
    function_impl_format,
    base_ability_format,
    BaseAbility,
    sys_response_class_format,
    sys_function_calling_format,
    sys_function_impl_format,
    exec_capture_output,
    format_prompt,
    format_prompt_jinja2,
)
from byzerllm.utils.ray_utils import cancel_placement_group, get_actor_info
from byzerllm.utils.json_repaire import repair_json_str
import byzerllm
import json
import importlib
import time
import functools
import inspect
import pydantic
import copy
import traceback
from enum import Enum
from loguru import logger

from byzerllm.utils.client.types import (
    Templates,
    Template,
    Role,
    LLMHistoryItem,
    LLMRequest,
    LLMFunctionCallResponse,
    LLMClassResponse,
    InferBackend,
    EventName,
    EventCallbackResult,
    EventCallback,
    LLMResponse,
    FintuneRequestExtra,
    FintuneRequest,
    ExecuteCodeResponse,
    LLMMetadata,
)


class ByzerLLM:

    def __init__(self, url: Optional[str] = None, **kwargs):
        self.url = url
        self.default_sys_conf = {
            "pythonMode": "ray",
            "maxConcurrency": 1,
            "num_gpus": 1,
            "masterMaxConcurrency": 1000,
            "workerMaxConcurrency": 1,
            "infer_backend": "transformers",
        }
        self.sys_conf = self.default_sys_conf.copy()
        self.sql_model = "context" in globals()

        self.verbose = kwargs.get("verbose", False)

        self.force_skip_context_length_check = False
        if "force_skip_context_length_check" in kwargs:
            self.force_skip_context_length_check = kwargs[
                "force_skip_context_length_check"
            ]

        self.mapping_auto_use_apply_chat_template = {}

        self.mapping_max_input_length = {}
        self.mapping_max_output_length = {}
        self.mapping_max_model_length = {}
        self.mapping_role_mapping = {}
        self.mapping_extra_generation_params = {}
        self.mapping_clean_func = {}

        self.mapping_function_calling_format_func = {}
        self.mapping_response_class_format_func = {}
        self.mapping_response_class_format_after_chat_func = {}
        self.mapping_impl_func_format_func = {}

        self.mapping_base_system_message = {}
        self.mapping_sys_response_class_format_func = {}
        self.mapping_sys_function_calling_format_func = {}
        self.mapping_sys_response_class_format_after_chat_func = {}
        self.mapping_sys_impl_func_format_func = {}

        self.func_impl_cache = {}
        self.meta_cache = {}

        self.byzer_engine_url = None
        if "byzer_engine_url" in kwargs:
            self.byzer_engine_url = kwargs["byzer_engine_url"]

        self.default_max_output_length = 4000
        if "default_max_output_length" in kwargs:
            self.default_max_output_length = kwargs["default_max_output_length"]

        self.default_model_name = None
        self.default_emb_model_name = None
        self.default_rerank_model_name = None
        self.default_role_mapping = {
            "user_role": "User:",
            "assistant_role": "Assistant:",
            "system_msg": "You are a helpful assistant. Think it over and answer the user question correctly.",
        }

        self.pin_model_worker_mapping = None

        if url is not None and self.sql_model:
            v = globals()
            self.context = v["context"]
            self.ray_context = RayContext.connect(v, self.url, **kwargs)
        else:
            self.context = PythonContext(0, [], self.sys_conf)
            self.context.have_fetched = True
            self.ray_context = self.context.rayContext

        self.event_callbacks: Dict[EventName, List[EventCallback]] = {}
        self.sub_clients = {}

    @property
    def metadata(self) -> LLMMetadata:
        meta = self.get_meta(model=self.default_model_name)
        return LLMMetadata(
            context_window=meta.get("max_model_len", 8192),
            num_output=meta.get("num_output", 256),
            is_chat_model=not meta.get("embedding_mode", False),
            is_function_calling_model=True,
            model_name=meta.get("model_name", self.default_model_name),
        )

    @staticmethod
    def from_default_model(model: str, auto_connect_cluster: bool = True) -> "ByzerLLM":
        if auto_connect_cluster:
            byzerllm.connect_cluster()
        llm = ByzerLLM()
        llm.setup_default_model_name(model)
        return llm

    def setup_sub_client(
        self, client_name: str, client: Optional["ByzerLLM"] = None
    ) -> "ByzerLLM":
        self.sub_clients[client_name] = client
        return self

    def get_sub_client(self, client_name: str) -> Optional["ByzerLLM"]:
        return self.sub_clients.get(client_name, None)

    def remove_sub_client(self, client_name: str) -> "ByzerLLM":
        if client_name in self.sub_clients:
            del self.sub_clients[client_name]
        return self

    def add_event_callback(
        self, event_name: EventName, callback: EventCallback
    ) -> None:
        self.event_callbacks.setdefault(event_name, []).append(callback)

    def _trigger_event(self, event_name: EventName, *args, **kwargs) -> Optional[Any]:
        if event_name in self.event_callbacks:
            for callback in self.event_callbacks[event_name]:
                continue_flag, value = callback(*args, **kwargs)
                if not continue_flag:
                    return value
        return None

    def setup_reset(self):
        self.sys_conf = self.default_sys_conf.copy()
        self.context.conf = self.sys_conf

    def setup_pin_model_worker_mapping(
        self, pin_model_worker_mapping: Dict[Any, int]
    ) -> "ByzerLLM":
        self.pin_model_worker_mapping = pin_model_worker_mapping
        return self

    def setup_load_balance_way(self, load_balance_way: str) -> "ByzerLLM":
        self.sys_conf["load_balance"] = load_balance_way
        return self

    def setup_default_model_name(self, model_name: str) -> "ByzerLLM":
        self.default_model_name = model_name
        return self

    def setup_default_emb_model_name(self, model_name: str) -> "ByzerLLM":
        self.default_emb_model_name = model_name
        return self

    def setup_default_re_rank_model_name(self, model_name: str) -> "ByzerLLM":
        self.default_rerank_model_name = model_name
        return self

    def setup(self, name: str, value: Any) -> "ByzerLLM":
        self.sys_conf[name] = value
        # update the context conf
        self.context.conf = self.sys_conf
        return self

    def setup_function_calling_format_func(self, model: str, func) -> "ByzerLLM":
        self.mapping_function_calling_format_func[model] = func
        return self

    def setup_response_class_format_func(self, model: str, func) -> "ByzerLLM":
        self.mapping_response_class_format_func[model] = func
        return self

    def setup_impl_func_format_func(self, model: str, func) -> "ByzerLLM":
        self.mapping_impl_func_format_func[model] = func
        return self

    def setup_response_class_format_after_chat_func(
        self, model: str, func
    ) -> "ByzerLLM":
        self.mapping_response_class_format_after_chat_func[model] = func
        return self

    def setup_base_system_messages(
        self, model: str, base_system_message: str
    ) -> "ByzerLLM":
        self.mapping_base_system_message[model] = base_system_message
        return self

    def setup_sys_response_class_format_func(self, model: str, func) -> "ByzerLLM":
        self.mapping_sys_response_class_format_func[model] = func
        return self

    def setup_sys_function_calling_format_func(self, model: str, func) -> "ByzerLLM":
        self.mapping_sys_function_calling_format_func[model] = func
        return self

    def setup_sys_response_class_format_after_chat_func(
        self, model: str, func
    ) -> "ByzerLLM":
        self.mapping_sys_response_class_format_after_chat_func[model] = func
        return self

    def setup_sys_impl_func_format_func(self, model: str, func) -> "ByzerLLM":
        self.mapping_sys_impl_func_format_func[model] = func
        return self

    def setup_infer_backend(self, backend: str) -> "ByzerLLM":
        self.sys_conf["infer_backend"] = backend

        if backend == InferBackend.VLLM or backend == InferBackend.DeepSpeed:
            self.sys_conf["masterMaxConcurrency"] = 1000
            self.sys_conf["workerMaxConcurrency"] = 100

        if backend == InferBackend.Transformers:
            self.sys_conf["masterMaxConcurrency"] = 1000
            self.sys_conf["workerMaxConcurrency"] = 1

        return self

    def setup_gpus_per_worker(self, num_gpus: int) -> "ByzerLLM":
        self.sys_conf["num_gpus"] = num_gpus
        return self

    def setup_cpus_per_worker(self, num_cpus: int) -> "ByzerLLM":
        self.sys_conf["num_cpus"] = num_cpus
        return self

    def setup_worker_concurrency(self, num: int) -> "ByzerLLM":
        self.sys_conf["workerMaxConcurrency"] = num
        return self

    def setup_num_workers(self, num_workers: int) -> "ByzerLLM":
        self.sys_conf["maxConcurrency"] = num_workers
        return self

    def setup_max_model_length(self, model: str, max_model_length: int) -> "ByzerLLM":
        self.mapping_max_model_length[model] = max_model_length
        return self

    def setup_max_input_length(self, model: str, max_input_length: int) -> "ByzerLLM":
        self.mapping_max_input_length[model] = max_input_length
        return self

    def setup_max_output_length(self, model: str, max_output_length: int) -> "ByzerLLM":
        self.mapping_max_output_length[model] = max_output_length
        return self

    def setup_role_mapping(
        self, model: str, role_mapping: Dict[str, str]
    ) -> "ByzerLLM":
        self.mapping_role_mapping[model] = role_mapping
        return self

    def setup_extra_generation_params(
        self, model: str, extra_generation_params: Dict[str, Any]
    ) -> "ByzerLLM":
        v = self.mapping_extra_generation_params.get(model, {})
        self.mapping_extra_generation_params[model] = {**v, **extra_generation_params}
        return self

    def setup_template(self, model: str, template: Union[Template, str]) -> "ByzerLLM":
        if template == "auto":
            meta = self.get_meta(model=model)

            is_saas_model = meta.get("model_deploy_type", None) == "saas"

            if is_saas_model:
                return self

            is_message_format = meta.get("message_format", False)

            if is_message_format:
                return self

            if "QWenLMHeadModel" in meta.get("architectures", []):
                self.setup_template(model, Templates.qwen())
                return self

            if not meta.get("support_chat_template", False):
                raise Exception(
                    f"The model({model}) is not support auto(apply chat template) for now."
                )

            self.mapping_auto_use_apply_chat_template[model] = True
            return self

        self.mapping_role_mapping[model] = template.role_mapping

        v = self.mapping_extra_generation_params.get(model, {})
        self.mapping_extra_generation_params[model] = {
            **v,
            **template.generation_config,
        }

        self.mapping_clean_func[model] = template.clean_func
        self.mapping_function_calling_format_func[model] = (
            template.function_calling_format_func
        )
        self.mapping_response_class_format_after_chat_func[model] = (
            template.response_class_format_after_chat_func
        )
        self.mapping_response_class_format_func[model] = (
            template.response_class_format_func
        )
        return self

    def sft(
        self,
        sft_name: str,
        local_data_dir_path: str,
        local_model_path: str,
        local_stage_path: str,
        pretrained_model_type: str,
        num_cpus: int,
        num_gpus: int,
        detached: bool = True,
        json_config: str = "{}",
        model_params: Dict[str, Any] = {},
        **kwargs,
    ):
        """
        finetune a pretrained model

        Args:
            sft_name (str): the uniq name of this finetune task
            local_data_dir_path (str): the local data dir path, which should contains `data.jsonl` file
            local_model_path (str): the local model path, which should contains `config.json` file
            local_stage_path (str): the local stage path which store the temp data and model
            pretrained_model_type (str): the pretrained model type, e.g. "sft/llama2","sft/baichuan"
            num_cpus (int): the number of cpus
            num_gpus (int): the number of gpus
            detached (bool, optional): whether to run this task in detached mode. Defaults to True.
            json_config (str, optional): the json config string. Defaults to "{}".
            model_params (Dict[str,Any], optional): the model params. Defaults to {}. The key should like this style `sft.int.logging_steps`, `sft.int.max_seq_length`
                                                    which contains the `sft` prefix and the type of the value.
        """
        train_params = {}
        train_params["name"] = sft_name
        train_params["data_dir"] = local_data_dir_path
        train_params["localModelDir"] = local_model_path
        train_params["pretrainedModelType"] = pretrained_model_type
        train_params["config"] = json_config
        train_params["detached"] = "true" if detached else "false"
        train_params["localPathPrefix"] = local_stage_path

        for k, v in model_params.items():
            train_params[k] = v

        sys_conf = {}
        sys_conf["num_gpus"] = num_gpus
        sys_conf["num_cpus"] = num_cpus

        r = self.raw_sft(train_params=train_params, sys_conf=sys_conf)
        if detached:
            return [i for i in r]
        return r

    def merge_lora(
        self,
        name: str,
        local_model_path: str,
        local_adpator_model_path: str,
        local_target_path: str,
    ):
        train_params = {}
        train_params["name"] = name
        train_params["modelNameOrPath"] = local_model_path
        train_params["adapterNameOrPath"] = local_adpator_model_path
        train_params["savePath"] = local_target_path
        self.raw_merge_lora(train_params=train_params, sys_conf={})
        return local_target_path

    def pretrain(
        self,
        name: str,
        local_data_dir_path: str,
        local_model_path: str,
        local_stage_path: str,
        pretrained_model_type: str,
        num_cpus: int,
        num_gpus: int,
        detached: bool = True,
        json_config: str = "{}",
        model_params: Dict[str, Any] = {},
        **kwargs,
    ):
        train_params = {}
        train_params["name"] = name
        train_params["localDataDir"] = local_data_dir_path
        train_params["localModelDir"] = local_model_path
        train_params["pretrainedModelType"] = pretrained_model_type
        train_params["deepspeedConfig"] = json_config
        train_params["detached"] = "true" if detached else "false"
        train_params["localPathPrefix"] = local_stage_path

        for k, v in model_params.items():
            train_params[k] = v

        sys_conf = {}
        sys_conf["num_gpus"] = num_gpus
        sys_conf["num_cpus"] = num_cpus

        r = self.raw_pretrain(train_params=train_params, sys_conf=sys_conf)
        if detached:
            return [i for i in r]
        return r

    def raw_sft(self, train_params: Dict[str, Any], sys_conf: Dict[str, Any] = {}):
        model_type = train_params["pretrainedModelType"].split("/")[-1]
        train_module = importlib.import_module(f"byzerllm.{model_type}")
        return train_module.sft_train([], train_params, sys_conf)

    def raw_pretrain(self, train_params: Dict[str, Any], sys_conf: Dict[str, Any] = {}):
        model_type = train_params["pretrainedModelType"][-1]
        train_module = importlib.import_module(f"byzerllm.{model_type}")
        return train_module.sfft_train([], train_params, sys_conf)

    def raw_merge_lora(self, train_params: Dict[str, Any], sys_conf: Dict[str, Any]):
        from byzerllm.utils.sft.merge_lora import merge_lora_to_base_model

        merge_lora_to_base_model([], train_params, sys_conf)

    def raw_deepspeed_to_huggingface(self, train_params: Dict[str, Any]):
        from byzerllm.utils.fulltune.pretrain.convert_to_transformers import convert

        convert(train_params, self.conf())

    def undeploy(self, udf_name: str, force: bool = False):
        import time

        try:
            model = ray.get_actor(udf_name)
            if not force:
                try:
                    meta = self.get_meta(model=udf_name)
                    if meta.get("backend", "") == "ray/vllm":
                        if "engine_placement_group_id" in meta:
                            cancel_placement_group(meta["engine_placement_group_id"])
                except Exception as inst:
                    pass
            ray.kill(model)
            if udf_name in self.meta_cache:
                del self.meta_cache[udf_name]
        except ValueError:
            pass
        time.sleep(3)

    def generate_instruction_from_history(
        self,
        model: str,
        conversations: List[Dict[str, str]],
        role_mapping: Dict[str, str] = {
            "user_role": "User:",
            "assistant_role": "Assistant:",
        },
    ):
        meta = self.get_meta(model=model)
        if self.mapping_auto_use_apply_chat_template.get(model, False) and meta.get(
            "support_chat_template", False
        ):
            return self.apply_chat_template(
                model, json.dumps(conversations, ensure_ascii=False)
            )

        new_his = []
        for item in conversations:
            if item["role"] == "system":
                value = item["content"]
                if "system_msg_func" in role_mapping:
                    value = role_mapping["system_msg_func"](
                        t=role_mapping["system_msg"], v=item["content"]
                    )
                new_his.append(value)
                continue

            if item["role"] == "user":
                value = f"{role_mapping['user_role']}{item['content']}"
                if "user_role_func" in role_mapping:
                    value = role_mapping["user_role_func"](
                        t=role_mapping["user_role"], v=item["content"]
                    )
                new_his.append(value)

            if item["role"] == "assistant":
                value = f"{role_mapping['assistant_role']}{item['content']}"
                if "user_role_func" in role_mapping:
                    value = role_mapping["assistant_role_func"](
                        t=role_mapping["assistant_role"], v=item["content"]
                    )
                new_his.append(value)

        if conversations[-1]["role"] == "user":
            new_his.append(f"{role_mapping['assistant_role']}")

        fin_ins = "\n".join(new_his)
        return fin_ins

    def is_model_exist(self, udf_name: str) -> bool:
        try:
            ray.get_actor(udf_name)
            return True
        except Exception as inst:
            return False

    def deploy(
        self,
        model_path: str,
        pretrained_model_type: str,
        udf_name: str,
        infer_params: Dict[str, Any],
    ):
        from byzerllm import common_init_model

        self.setup("UDF_CLIENT", udf_name)

        infer_backend = self.sys_conf["infer_backend"]

        if (
            infer_backend == InferBackend.VLLM
            or infer_backend == InferBackend.DeepSpeed
        ):
            if pretrained_model_type != "custom/auto":
                raise ValueError(
                    f"Backend({infer_backend}) is set. the pretrained_model_type should be `custom/auto`"
                )

        model_type = pretrained_model_type

        if pretrained_model_type.startswith("saas/"):
            model_type = pretrained_model_type.split("/")[-1]

            infer_module = importlib.import_module(f"byzerllm.saas.{model_type}")
            from byzerllm.utils.text_generator import simple_predict_func

            def init_model(
                model_refs: List[ClientObjectRef], conf: Dict[str, str]
            ) -> Any:
                from byzerllm import consume_model

                consume_model(conf)
                infer = infer_module.CustomSaasAPI(infer_params)
                return (infer, None)

            UDFBuilder.build(self.ray_context, init_model, simple_predict_func)
            return self.get_meta(model=udf_name)

        if pretrained_model_type == "bark":
            from byzerllm.bark.bark_voice import (
                build_void_infer,
                ZH_SPEAKER,
                EN_SPEAKER,
            )

            def init_model(
                model_refs: List[ClientObjectRef], conf: Dict[str, str]
            ) -> Any:
                infer = build_void_infer(
                    model_dir=model_path,
                    tokenizer_dir=f"{model_path}/pretrained_tokenizer",
                )
                return infer

            def predict_func(model, v):
                data = [json.loads(item) for item in v]
                results = [
                    {
                        "predict": model.text_to_voice(item["instruction"]).tolist(),
                        "labels": "",
                    }
                    for item in data
                ]
                return {"value": [json.dumps(results, ensure_ascii=False, indent=4)]}

            UDFBuilder.build(self.ray_context, init_model, predict_func)
            return self.get_meta(model=udf_name)

        # we put in this place so it only take effect for private model
        self.mapping_max_output_length[udf_name] = 4000

        if pretrained_model_type.startswith("custom/"):
            model_type = pretrained_model_type.split("/")[-1]

        predict_func = "simple_predict_func"
        if model_type == "chatglm2":
            predict_func = "chatglm_predict_func"

        infer_module = importlib.import_module(f"byzerllm.{model_type}")
        predict_module = importlib.import_module(f"byzerllm.utils.text_generator")

        def init_model(model_refs: List[ClientObjectRef], conf: Dict[str, str]) -> Any:
            common_init_model(model_refs, conf, model_path, is_load_from_local=True)
            model = infer_module.init_model(model_path, infer_params, conf)
            return model

        UDFBuilder.build(
            self.ray_context, init_model, getattr(predict_module, predict_func)
        )
        return self.get_meta(model=udf_name)

    def get_meta(self, model: str, llm_config: Dict[str, Any] = {}):
        if not model and not self.default_model_name:
            raise Exception("model name is required")

        if not model:
            model = self.default_model_name

        if model in self.meta_cache:
            return self.meta_cache[model]

        default_config = self.mapping_extra_generation_params.get(model, {})

        v = [{"instruction": "", "meta": True, **{**default_config, **llm_config}}]
        res = self._query(model, v)

        t = [
            LLMResponse(
                output=item["predict"],
                metadata=item.get("metadata", {}),
                input=item["input"],
            )
            for item in res
        ]

        res = {}
        if len(t) != 0 and len(t[0].output) != 0:
            res = t[0].output[0]

        self.meta_cache[model] = res
        return self.meta_cache[model]

    def tokenize(
        self, model: str, s: str, llm_config: Dict[str, Any] = {}
    ) -> List[str]:

        if not model and not self.default_model_name:
            raise Exception("model name is required")

        if not model:
            model = self.default_model_name

        default_config = self.mapping_extra_generation_params.get(model, {})

        v = [{"instruction": s, "tokenizer": True, **{**default_config, **llm_config}}]
        res = self._query(model, v)
        return [
            LLMResponse(
                output=item["predict"],
                metadata=item.get("metadata", {}),
                input=item["input"],
            )
            for item in res
        ]

    def apply_chat_template(self, model: str, s: str, llm_config: Dict[str, Any] = {}):
        if not model and not self.default_model_name:
            raise Exception("model name is required")

        if not model:
            model = self.default_model_name

        default_config = self.mapping_extra_generation_params.get(model, {})
        v = [
            {
                "instruction": s,
                "apply_chat_template": True,
                **{**default_config, **llm_config},
            }
        ]
        res = self._query(model, v)

        t = [
            LLMResponse(
                output=item["predict"],
                metadata=item.get("metadata", {}),
                input=item["input"],
            )
            for item in res
        ]
        return t[0].output

    def emb_query(self, v: str, model: str = None):
        return self.emb(model=model, request=LLMRequest(instruction=v))

    def emb(self, model, request: LLMRequest, extract_params: Dict[str, Any] = {}):

        if not model and not self.default_emb_model_name:
            raise Exception("model name is required")

        if not model:
            model = self.default_emb_model_name

        default_config = self.mapping_extra_generation_params.get(model, {})

        if isinstance(request, list):
            request = LLMRequest(instruction=request)

        if isinstance(request.instruction, str):
            v = [
                {
                    "instruction": request.instruction,
                    "embedding": True,
                    "max_length": request.max_length,
                    "top_p": request.top_p,
                    "temperature": request.temperature,
                    **default_config,
                    **extract_params,
                }
            ]
        else:
            v = [
                {
                    "instruction": x,
                    "embedding": True,
                    "max_length": request.max_length,
                    "top_p": request.top_p,
                    "temperature": request.temperature,
                    **default_config,
                    **extract_params,
                }
                for x in request.instruction
            ]
        res = self._query(model, v)

        return [
            LLMResponse(
                output=item["predict"],
                metadata=item.get("metadata", {}),
                input=item["input"],
            )
            for item in res
        ]

    def emb_rerank(
        self,
        model: str = None,
        sentence_pairs: Union[List[Tuple[str, str]], Tuple[str, str]] = [],
        extract_params: Dict[str, Any] = {},
    ) -> Union[Tuple[Tuple[str, str], float], List[Tuple[Tuple[str, str], float]]]:

        if not model and not self.default_rerank_model_name:
            raise Exception("rerank model name is required")

        if not sentence_pairs or len(sentence_pairs) == 0:
            raise Exception("rerank rerank param sentence_pairs is required")

        if not model:
            model = self.default_rerank_model_name

        default_config = self.mapping_extra_generation_params.get(model, {})

        v = [
            {
                "instruction": sentence_pairs,
                "embedding": True,
                "embed_rerank": True,
                **default_config,
                **extract_params,
            }
        ]
        res = self._query(model, v)

        return [
            LLMResponse(
                output=item["predict"],
                metadata=item.get("metadata", {}),
                input=item["input"],
            )
            for item in res
        ]

    def _generate_ins(
        self, model: str, request: LLMRequest, role_mapping: Dict[str, str]
    ):
        if not role_mapping["user_role"]:
            return request.instruction

        sys_msg = role_mapping["system_msg"]
        if "system_msg_func" in role_mapping:
            sys_msg = "You are a helpful assistant. Think it over and answer the user question correctly."

        conversations = [{"role": "system", "content": sys_msg}]
        # conversations += [{"role":item.role,"content":item.content} for item in request.extra_params.history]

        conversations += self._to_openai_format(request=request)

        final_ins = self.generate_instruction_from_history(
            model, conversations, role_mapping
        )

        return final_ins

    def _to_openai_format(self, request: LLMRequest):
        conversations = []
        if isinstance(request.instruction, str):
            conversations += [{"role": "user", "content": request.instruction}]
        else:
            conversations += [
                {"role": "user", "content": x} for x in request.instruction
            ]
        return conversations

    def execute_function_calling(
        self, response: LLMResponse, tools: List[Callable], func_params: Dict[str, Any]
    ) -> LLMFunctionCallResponse:

        r = LLMFunctionCallResponse(
            response=response, values=[], metadata={"reason": ""}
        )

        is_json = False
        try:
            json.loads(response.output)
            is_json = True
        except Exception as inst:
            pass

        code = response.output

        if not is_json:
            if code.strip().startswith("```json"):
                index = code.rfind("```")
                if index != -1:
                    code = code.strip()[7 : index - 1]
                    try:
                        json.loads(code)
                        is_json = True
                    except Exception as inst:
                        pass

        if not is_json:
            codes = code_utils.extract_code(response.output)
            if len(codes) == 0:
                r.metadata["reason"] = "No json block found"
                return r

            lang, code = codes[-1]

            if lang != "json":
                r.metadata["reason"] = "No json block found"
                return r

        try:
            temp = json.loads(code)
            if isinstance(temp, list):
                temp = temp[-1]
            ms = FunctionCallList.parse_obj(temp)
        except Exception as inst:
            r.metadata["reason"] = str(inst) + "\n" + traceback.format_exc()
            return r

        _func_maps = dict([(t.__name__, t) for t in tools])

        if func_params is None:
            func_params = {}

        try:
            r.metadata["selected_functions"] = []
            for m in ms.tool_calls:
                if m.function.name in _func_maps:
                    r.metadata["selected_functions"].append(m.function.name)
                    r.values.append(
                        _func_maps[m.function.name](
                            **m.function.arguments, **func_params
                        )
                    )
        except Exception as inst:
            r.metadata["reason"] = str(inst) + "\n" + traceback.format_exc()

        return r

    def execute_generate_func(
        self,
        func_name: str,
        impl_func_params: Optional[Dict[str, Any]],
        response: LLMResponse,
        response_class: pydantic.BaseModel,
    ) -> LLMClassResponse:

        r = LLMClassResponse(response=response, value=None, metadata={"reason": ""})

        is_python_code = False
        if code_utils.infer_lang(response.output) == "python":
            is_python_code = True

        code = response.output

        if not is_python_code:
            codes = code_utils.extract_code(response.output)

            if len(codes) == 0:
                r.metadata["reason"] = "No Python block found"
                return r

            lang, code = codes[-1]

            if lang != "python":
                r.metadata["reason"] = "No Python block found"
                return r

        (status, output, variables) = exec_capture_output(code, {func_name: True})
        if status != 0:
            r.metadata["reason"] = output
            return r

        try:
            if impl_func_params is None:
                impl_func_params = {}
            res_json = variables[func_name](**impl_func_params)
            r.metadata["raw_func"] = code
            r.metadata["func"] = variables[func_name]
            if isinstance(res_json, str):
                res_json = json.loads(res_json)
            r.value = response_class.parse_obj(res_json)
        except Exception as inst:
            r.metadata["reason"] = str(inst) + "\n" + traceback.format_exc()
            return r

        return r

    def execute_response_format(
        self, response: LLMResponse, response_class: pydantic.BaseModel
    ):

        r = LLMClassResponse(response=response, value=None, metadata={"reason": ""})
        is_json = False
        try:
            json.loads(response.output)
            is_json = True
        except Exception as inst:
            pass

        code = response.output

        if not is_json:
            if code.strip().startswith("```json"):
                index = code.rfind("```")
                if index != -1:
                    code = code.strip()[7 : index - 1]
                    try:
                        json.loads(code)
                        is_json = True
                    except Exception as inst:
                        pass

        if not is_json:
            codes = code_utils.extract_code(response.output)
            if len(codes) == 0:
                r.metadata["reason"] = "No json block found"
                return r

            lang, code = codes[-1]

            if lang != "json":
                r.metadata["reason"] = "No json block found"
                return r

        try:
            try:
                obj = json.loads(code)
            except Exception as inst:
                print(
                    "Fail to parse json. Error:\n"
                    + str(inst)
                    + "\n"
                    + traceback.format_exc(),
                    flush=True,
                )
                obj = json.loads(repair_json_str(code))
            ms = response_class.parse_obj(obj)
        except Exception as inst:
            r.metadata["reason"] = str(inst) + "\n" + traceback.format_exc()
            return r

        r.value = ms

        return r

    def require_template(self, model: str):
        meta = self.get_meta(model=model)
        is_saas_model = meta.get("model_deploy_type", None) == "saas"
        is_message_format = meta.get("message_format", False)
        support_chat_template = meta.get("support_chat_template", False)
        return not is_saas_model and not is_message_format and not support_chat_template

    def abort(self, request_id: str, model: Optional[str] = None):
        if not model and not self.default_model_name:
            raise Exception("model name is required")
        if not model:
            model = self.default_model_name

        meta = self.get_meta(model=model)
        if meta.get("backend", None) != "ray/vllm":
            raise Exception("abort only support ray/vllm backend")

        self.chat_oai(
            conversations=[{"role": "user", "content": f"{request_id}"}],
            llm_config={"gen.request_id": request_id, "gen.abort": True},
        )

    def chat_oai(
        self,
        conversations,
        tools: List[Union[Callable, str]] = [],
        tool_choice: Optional[Union[Callable, str]] = None,
        execute_tool: bool = False,
        impl_func: Optional[Callable] = None,
        execute_impl_func: bool = False,
        impl_func_params: Optional[Dict[str, Any]] = None,
        func_params: Optional[Dict[str, Any]] = None,
        response_class: Optional[Union[pydantic.BaseModel, str]] = None,
        response_after_chat: Optional[Union[pydantic.BaseModel, str]] = False,
        enable_default_sys_message: bool = True,
        model: Optional[str] = None,
        role_mapping=None,
        llm_config: Dict[str, Any] = {},
        only_return_prompt: bool = False,
    ) -> Union[
        List[LLMResponse], List[LLMFunctionCallResponse], List[LLMClassResponse]
    ]:

        if not self.default_model_name and not model:
            raise Exception(
                "Use llm.setup_default_model_name to setup default model name or setup the model parameter"
            )

        if not model:
            model = self.default_model_name

        if role_mapping is None:
            role_mapping = self.mapping_role_mapping.get(
                model, self.default_role_mapping
            )

        if response_class and (tools or tool_choice):
            raise Exception(
                "function calling is enabled,response_class should not be set."
            )

        if impl_func and not response_class:
            raise Exception("impl_func is enabled,response_class should be set.")

        if isinstance(conversations, str):
            conversations = [{"role": "user", "content": conversations}]

        if enable_default_sys_message:
            first_message = conversations[0]
            base_abilities = []
            if response_class:
                base_abilities.append(BaseAbility.RESPONSE_WITH_CLASS)
            if impl_func:
                base_abilities.append(BaseAbility.FUNCTION_IMPL)
            if tools or tool_choice:
                base_abilities.append(BaseAbility.FUNCTION_CALLING)

            if base_abilities and first_message["role"] == "user":
                conversations.insert(
                    0,
                    {
                        "role": "system",
                        "content": self.mapping_base_system_message.get(
                            model, base_ability_format(base_abilities=base_abilities)
                        ),
                    },
                )

            if first_message["role"] == "system":
                first_message[
                    "content"
                ] = f"""{self.mapping_base_system_message.get(model,base_ability_format(base_abilities=base_abilities))}
{first_message["content"]}"""

        meta = self.get_meta(model=model)
        is_saas_model = meta.get("model_deploy_type", None) == "saas"
        is_message_format = meta.get("message_format", False)

        temp_conversations = copy.deepcopy(conversations)
        last_message = temp_conversations[-1]

        # function calling
        if tools or tool_choice:
            f = (
                self.mapping_function_calling_format_func.get(
                    model, function_calling_format
                )
                if not enable_default_sys_message
                else self.mapping_sys_function_calling_format_func.get(
                    model, sys_function_calling_format
                )
            )
            last_message["content"] = f(last_message["content"], tools, tool_choice)

        # implement function and the function should return a response class
        elif impl_func and response_class:
            f = (
                self.mapping_impl_func_format_func.get(model, function_impl_format)
                if not enable_default_sys_message
                else self.mapping_sys_impl_func_format_func.get(
                    model, sys_function_impl_format
                )
            )
            last_message["content"] = f(
                last_message["content"], impl_func, cls=response_class
            )

        # generate response class
        elif response_class and not response_after_chat:
            f = (
                self.mapping_response_class_format_func.get(
                    model, response_class_format
                )
                if not enable_default_sys_message
                else self.mapping_sys_response_class_format_func.get(
                    model, sys_response_class_format
                )
            )
            last_message["content"] = f(last_message["content"], cls=response_class)

        if is_saas_model or is_message_format:
            final_ins = last_message["content"]
            history = []
            for item in temp_conversations[:-1]:
                # clean metadata field in conversation
                # which may used by agent.
                if "metadata" in item:
                    del item["metadata"]
                history.append(item)

        else:
            final_ins = self.generate_instruction_from_history(
                model, temp_conversations, role_mapping
            )
            history = []

        default_config = self.mapping_extra_generation_params.get(model, {})

        if self.get_max_output_length(model) > 0:
            default_config["max_length"] = self.get_max_output_length(model)

        v = [
            {
                "instruction": final_ins,
                "history": history,
                **default_config,
                **llm_config,
            }
        ]

        if only_return_prompt:
            responses = [
                LLMResponse(output="", metadata=item, input=item["instruction"])
                for item in v
            ]
            if response_class or response_after_chat:
                new_responses = []
                for response in responses:
                    temp = LLMClassResponse(
                        response=response,
                        value=response,
                        metadata={"reason": "Only return prompt"},
                    )
                    new_responses.append(temp)
                return new_responses
            return responses

        res = self._query(model, v)
        clean_func = self.mapping_clean_func.get(model, lambda s: s)

        responses = [
            LLMResponse(
                output=clean_func(item["predict"]),
                metadata=item.get("metadata", {}),
                input=item["input"],
            )
            for item in res
        ]

        ## handle impl_func response
        if impl_func and response_class and execute_impl_func:
            final_result = []
            for response in responses:
                final_result.append(
                    self.execute_generate_func(
                        func_name=impl_func.__name__,
                        impl_func_params=impl_func_params or func_params,
                        response=response,
                        response_class=response_class,
                    )
                )
            return final_result

        if impl_func and response_class:
            return responses

        ## handle response_class response
        temp_result = responses
        if response_class and response_after_chat:
            temp_result = []
            f = self.mapping_response_class_format_after_chat_func.get(
                model, response_class_format_after_chat
            )
            for response in responses:
                new_conversations = temp_conversations + [
                    {"content": response.output, "role": "assistant"},
                    {"content": f(response_class), "role": "user"},
                ]
                temp_result.append(
                    self.chat_oai(
                        new_conversations,
                        role_mapping=role_mapping,
                        llm_config=llm_config,
                    )[0]
                )

        if response_class:
            final_result = []
            for response in temp_result:
                final_result.append(
                    self.execute_response_format(
                        response=response, response_class=response_class
                    )
                )
            return final_result

        ## handle function calling response
        if execute_tool:
            final_result = []
            for response in responses:
                final_result.append(
                    self.execute_function_calling(
                        response=response, tools=tools, func_params=func_params
                    )
                )

            return final_result

        return responses

    def stream_chat_oai(
        self,
        conversations,
        model: Optional[str] = None,
        role_mapping=None,
        delta_mode: bool = False,
        llm_config: Dict[str, Any] = {},
    ):

        if not model:
            model = self.default_model_name

        meta = self.get_meta(model=model)
        if not meta.get("support_stream", False):
            raise Exception(f"The model({model}) is not support stream chat for now.")

        v = self.chat_oai(
            conversations,
            model=model,
            role_mapping=role_mapping,
            llm_config={**llm_config, **{"generation.stream": True}},
        )
        request_id = v[0].metadata["request_id"]
        stream_server_type = v[0].metadata.get("stream_server", "VLLM_STREAM_SERVER")
        server = ray.get_actor(stream_server_type)

        pre_generated_text = None
        while True:
            final_output = ray.get(server.get_item.remote(request_id))
            if isinstance(final_output, str):
                time.sleep(0.01)
                continue

            if final_output is None:
                break

            if stream_server_type == "BlockBinaryStreamServer":
                binary_data = final_output.outputs[0].text
                yield (binary_data, final_output.outputs[0].metadata)
            else:
                text_outputs = final_output.outputs
                clean_func = self.mapping_clean_func.get(model, lambda s: s)
                generated_text = text_outputs[0].text
                if (
                    pre_generated_text is not None
                    and generated_text == pre_generated_text
                ):
                    continue

                if delta_mode and pre_generated_text is not None:
                    s = generated_text[len(pre_generated_text) :]
                else:
                    s = generated_text
                pre_generated_text = generated_text
                yield (clean_func(s), text_outputs[0].metadata)

    async def async_stream_chat_oai(
        self,
        conversations,
        role_mapping=None,
        model: Optional[str] = None,
        delta_mode: bool = False,
        llm_config: Dict[str, Any] = {},
    ):

        if not model:
            model = self.default_model_name

        meta = self.get_meta(model=model)
        if not meta.get("support_stream", False):
            raise Exception(f"The model({model}) is not support stream chat for now.")

        v = self.chat_oai(
            conversations,
            model=model,
            role_mapping=role_mapping,
            llm_config={**llm_config, **{"generation.stream": True}},
        )
        request_id = v[0].metadata["request_id"]
        stream_server_type = v[0].metadata.get("stream_server", "VLLM_STREAM_SERVER")
        server = ray.get_actor(stream_server_type)

        pre_generated_text = None
        while True:
            final_output = await server.get_item.remote(request_id)
            if isinstance(final_output, str):
                time.sleep(0.01)
                continue

            if final_output is None:
                break

            if stream_server_type == "BlockBinaryStreamServer":
                binary_data = final_output.outputs[0].text
                yield (binary_data, final_output.outputs[0].metadata)
            else:
                text_outputs = final_output.outputs
                clean_func = self.mapping_clean_func.get(model, lambda s: s)
                generated_text = text_outputs[0].text
                if (
                    pre_generated_text is not None
                    and generated_text == pre_generated_text
                ):
                    continue

                if delta_mode and pre_generated_text is not None:
                    s = generated_text[len(pre_generated_text) :]
                else:
                    s = generated_text
                pre_generated_text = generated_text
                yield (clean_func(s), text_outputs[0].metadata)

    def clear_impl_cache(
        self,
        model: Optional[str] = None,
        full_func_name: Optional[str] = None,
        instruction: Optional[str] = None,
    ):
        if model is None and full_func_name is None and instruction is None:
            self.func_impl_cache = {}

        if model is not None and full_func_name is not None and instruction is None:
            raise Exception("instruction is required")

        if model is not None:
            instruction = "" if not instruction else instruction
            full_func_name = "" if not full_func_name else full_func_name

            key = f"{model}_{instruction}_{full_func_name}"
            for k in list(self.func_impl_cache.keys()):
                if k.startswith(key):
                    del self.func_impl_cache[k]
            return self

        if full_func_name is not None:
            instruction = "" if not instruction else instruction
            model = "" if not model else model
            key = f"{model}_{instruction}_{full_func_name}"
            for k in list(self.func_impl_cache.keys()):
                if k.endswith(key):
                    del self.func_impl_cache[k]
            return self

    def prompt(
        self,
        model: Optional[str] = None,
        render: Optional[str] = "jinja2",
        check_result: bool = False,
        options: Dict[str, Any] = {},
        return_origin_response: bool = False,
        marker: Optional[str] = None,
    ):
        if model is None:
            if "model" in options:
                model = options.pop("model")
            else:
                model = self.default_model_name

        def is_instance_of_generator(v):
            from typing import Generator, get_origin, get_args
            import collections                        
            if get_origin(v) is collections.abc.Generator:
                args = get_args(v)
                if args == (str, type(None), type(None)):
                    return True
            return False

        def _impl(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                signature = inspect.signature(func)
                arguments = signature.bind(*args, **kwargs)
                arguments.apply_defaults()
                input_dict = {}
                for param in signature.parameters:
                    input_dict.update({param: arguments.arguments[param]})

                if "self" in input_dict:
                    instance = input_dict.pop("self")
                    new_input_dic = func(instance, **input_dict)
                    if new_input_dic and not isinstance(new_input_dic, dict):
                        raise TypeError(
                            f"Return value of {func.__name__} should be a dict"
                        )
                    if new_input_dic:
                        input_dict = {**input_dict, **new_input_dic}
                else:
                    new_input_dic = func(**input_dict)
                    if new_input_dic and not isinstance(new_input_dic, dict):
                        raise TypeError(
                            f"Return value of {func.__name__} should be a dict"
                        )
                    if new_input_dic:
                        input_dict = {**input_dict, **new_input_dic}

                if render == "jinja2" or render == "jinja":
                    prompt_str = format_prompt_jinja2(func, **input_dict)
                else:
                    prompt_str = format_prompt(func, **input_dict)

                if marker:
                    prompt_str = f"{prompt_str}\n\n{marker}"
                                
                if is_instance_of_generator(signature.return_annotation):
                    temp_options = {**{"delta_mode":True}, **options}
                    t = self.stream_chat_oai(
                        conversations=[{"role": "user", "content": prompt_str}],
                        model=model,                        
                        **temp_options,
                    )
                    if return_origin_response:
                        return t
                    return (item[0] for item in t)

                if issubclass(signature.return_annotation, pydantic.BaseModel):
                    response_class = signature.return_annotation
                    t = self.chat_oai(
                        model=model,
                        conversations=[{"role": "user", "content": prompt_str}],
                        response_class=response_class,
                        impl_func_params=input_dict,
                        **options,
                    )
                    if return_origin_response:
                        return t
                    r: LLMClassResponse = t[0]
                    if r.value is None and check_result:
                        logger.warning(
                            f"""
                                {func.__name__} return None.
                                metadata:
                                {r.metadata}
                                response:
                                {r.response}
                            """
                        )
                    return r.value
                elif issubclass(signature.return_annotation, str):
                    t = self.chat_oai(
                        model=model,
                        conversations=[{"role": "user", "content": prompt_str}],
                        **options,
                    )
                    if return_origin_response:
                        return t
                    return t[0].output
                else:
                    raise Exception(
                        f"{func.__name__} should return a pydantic model or string"
                    )

            return wrapper

        return _impl

    def response(
        self,
        instruction: Optional[str] = None,
        model: Optional[str] = None,
        verbose: Optional[bool] = None,
    ):
        if model is None:
            model = self.default_model_name
        if instruction is None:
            instruction = ""

        if verbose is None:
            verbose = self.verbose

        def _impl(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                signature = inspect.signature(func)
                arguments = signature.bind(*args, **kwargs)
                arguments.apply_defaults()
                input_dict = {}
                for param in signature.parameters:
                    input_dict.update({param: arguments.arguments[param]})

                if len(input_dict.keys()) != 1:
                    raise Exception(
                        "response function should have only one parameter which type should be string"
                    )

                if issubclass(signature.return_annotation, pydantic.BaseModel):
                    response_class = signature.return_annotation
                else:
                    raise Exception("impl function should return a pydantic model")

                start_time = time.monotonic()

                t = self.chat_oai(
                    model=model,
                    conversations=[
                        {"role": "user", "content": list(input_dict.values())[0]}
                    ],
                    response_class=response_class,
                    impl_func_params=input_dict,
                )

                r: LLMClassResponse = t[0]

                if verbose:
                    print(
                        f"""cost {time.monotonic() - start_time} seconds""", flush=True
                    )

                return r.value

            return wrapper

        return _impl

    def impl(
        self,
        instruction: Optional[str] = None,
        model: Optional[str] = None,
        verbose: Optional[bool] = None,
        skip_cache: bool = False,
    ):
        if model is None:
            model = self.default_model_name
        if instruction is None:
            instruction = ""

        if verbose is None:
            verbose = self.verbose

        def _impl(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):

                key = f"{model}_{instruction}_{func.__module__}.{func.__name__}"
                signature = inspect.signature(func)
                arguments = signature.bind(*args, **kwargs)
                arguments.apply_defaults()

                if issubclass(signature.return_annotation, pydantic.BaseModel):
                    response_class = signature.return_annotation
                else:
                    raise Exception("impl function should return a pydantic model")

                if not skip_cache and key in self.func_impl_cache:
                    if verbose:
                        print(f""" {key} in cache, skip impl function""")
                    return response_class.parse_obj(
                        self.func_impl_cache[key](*args, **kwargs)
                    )

                input_dict = {}
                for param in signature.parameters:
                    input_dict.update({param: arguments.arguments[param]})

                start_time = time.monotonic()

                t = self.chat_oai(
                    model=model,
                    conversations=[{"role": "user", "content": instruction}],
                    impl_func=func,
                    response_class=response_class,
                    execute_impl_func=True,
                    impl_func_params=input_dict,
                )

                r: LLMClassResponse = t[0]

                if verbose:
                    print(
                        f"""Generate code for {key}: 
```python
{r.metadata["raw_func"]}
``` 
cost {time.monotonic() - start_time} seconds                     
""",
                        flush=True,
                    )

                if not skip_cache and key not in self.func_impl_cache:
                    self.func_impl_cache[key] = r.metadata["func"]

                return r.value

            return wrapper

        return _impl

    def raw_chat(
        self,
        model,
        request: Union[LLMRequest, str],
        extract_params: Dict[str, Any] = {},
    ) -> List[LLMResponse]:
        if isinstance(request, str):
            request = LLMRequest(instruction=request)

        return self.chat(model, request, extract_params)

    def chat(
        self,
        model,
        request: Union[LLMRequest, str],
        extract_params: Dict[str, Any] = {},
    ) -> List[LLMResponse]:
        if not model and not self.default_model_name:
            raise Exception("model name is required")

        if not model:
            model = self.default_model_name

        default_config = self.mapping_extra_generation_params.get(model, {})

        default_role_mapping = self.mapping_role_mapping.get(
            model, self.default_role_mapping
        )

        if isinstance(request, str):
            request = LLMRequest(instruction=request)

        if isinstance(request.instruction, str):

            final_input = self._generate_ins(model, request, default_role_mapping)

            v = [
                {
                    "instruction": final_input,
                    "max_length": request.max_length,
                    "top_p": request.top_p,
                    "temperature": request.temperature,
                    **default_config,
                    **extract_params,
                }
            ]
        else:
            v = []
            for x in request.instruction:

                new_request = LLMRequest(
                    instruction=x,
                    embedding=request.embedding,
                    max_length=request.max_length,
                    top_p=request.top_p,
                    temperature=request.temperature,
                )

                final_input = self._generate_ins(
                    model, new_request, default_role_mapping
                )

                v.append(
                    {
                        "instruction": final_input,
                        "max_length": request.max_length,
                        "top_p": request.top_p,
                        "temperature": request.temperature,
                        **default_config,
                        **extract_params,
                    }
                )
        res = self._query(model, v)
        clean_func = self.mapping_clean_func.get(model, lambda s: s)
        return [
            LLMResponse(
                output=clean_func(item["predict"]),
                metadata=item.get("metadata", {}),
                input=item["input"],
            )
            for item in res
        ]

    def apply_sql_func(
        self,
        sql: str,
        data: List[Dict[str, Any]],
        owner: str = "admin",
        url: str = "http://127.0.0.1:9003/model/predict",
    ):
        if self.byzer_engine_url and url == "http://127.0.0.1:9003/model/predict":
            url = self.byzer_engine_url
        res = self._rest_byzer_engine(sql, data, owner, url)
        return res

    def _rest_byzer_script(
        self, sql: str, owner: str, url: str = "http://127.0.0.1:9003/run/script"
    ):
        import requests
        import json

        data = {
            "sessionPerUser": "true",
            "sessionPerRequest": "true",
            "owner": owner,
            "sql": sql,
            "includeSchema": True,
        }
        response = requests.post(url, data=data)

        if response.status_code != 200:
            raise Exception(
                f"{self.url} status:{response.status_code} content: {response.text} request: json/{json.dumps(data,ensure_ascii=False)}"
            )
        res = json.loads(response.text)
        return res

    def _rest_byzer_engine(
        self, sql: str, table: List[Dict[str, Any]], owner: str, url: str
    ):
        import requests
        import json

        data = {
            "sessionPerUser": "true",
            "sessionPerRequest": "true",
            "owner": owner,
            "dataType": "row",
            "sql": sql,
            "data": json.dumps(table, ensure_ascii=False),
        }
        response = requests.post(url, data=data)

        if response.status_code != 200:
            raise Exception(
                f"{self.url} status:{response.status_code} content: {response.text} request: json/{json.dumps(data,ensure_ascii=False)}"
            )
        res = json.loads(response.text)
        return res[0]

    def get_max_model_length(self, model: str):
        return self.mapping_max_model_length.get(model, None)

    def get_max_output_length(self, model: str):
        return self.mapping_max_output_length.get(model, self.default_max_output_length)

    def get_max_input_length(self, model: str):
        return self.mapping_max_input_length.get(model, None)

    def _query(self, model: str, input_value: List[Dict[str, Any]]):

        try:
            from byzerllm.utils.nontext import Image, Audio

            for v in input_value:
                s = v["instruction"]
                image = Image(s)
                if image.has_image():
                    c = image.to_content()
                    v["instruction"] = json.dumps(c, ensure_ascii=False)

                audio = Audio(s)
                if audio.has_audio():
                    c = audio.to_content()
                    v["instruction"] = json.dumps(c, ensure_ascii=False)                

        except Exception as inst:
            pass        

        event_result = self._trigger_event(
            EventName.BEFORE_CALL_MODEL, self, model, input_value
        )
        if event_result is not None:
            return event_result

        udf_master = ray.get_actor(model)

        try:
            new_input_value = [json.dumps(x, ensure_ascii=False) for x in input_value]
        except Exception as inst:
            raise Exception(
                f"input_value should be json serializable, got {input_value}"
            )

        if self.verbose:
            print(f"Send to model[{model}]:{new_input_value}")
        index = -1
        try:
            worker_id = -1
            if self.pin_model_worker_mapping:
                if input_value[0].get("embedding", False):
                    worker_id = self.pin_model_worker_mapping.get("embedding", -1)
                elif input_value[0].get("tokenizer", False):
                    worker_id = self.pin_model_worker_mapping.get("tokenizer", -1)
                elif input_value[0].get("apply_chat_template", False):
                    worker_id = self.pin_model_worker_mapping.get(
                        "apply_chat_template", -1
                    )
                elif input_value[0].get("meta", False):
                    worker_id = self.pin_model_worker_mapping.get("meta", -1)

            [index, worker] = ray.get(udf_master.get.remote(worker_id))
            res = ray.get(worker.async_apply.remote(new_input_value))

            event_result = self._trigger_event(
                EventName.AFTER_CALL_MODEL, self, model, json.loads(res["value"][0])
            )
            if event_result is not None:
                return event_result

            return json.loads(res["value"][0])
        except Exception as inst:
            raise inst
        finally:
            if index != -1:
                ray.get(udf_master.give_back.remote(index))



from typing import List, Optional, Tuple,Any
from transformers import AutoTokenizer, AutoModelForCausalLM,BitsAndBytesConfig
import ray
import torch
import deepspeed
import os
from ray.air.util.torch_dist import (
    ActorHandle,
    _get_node_and_gpu_ids,
    _init_torch_distributed,
    get_address_and_port,
)
from ray.train.constants import DEFAULT_NCCL_SOCKET_IFNAME

class ParallelConfig:
    """Configuration for the distributed execution.

    Args:
        pipeline_parallel_size: Number of pipeline parallel groups.
        tensor_parallel_size: Number of tensor parallel groups.        
    """

    def __init__(
        self,
        num_workers:int,
        model_dir:str,        
        backend: str = "nccl",              
    ) -> None:
        self.world_size = num_workers
        self.model_dir = model_dir
        self.backend = backend
    

DeviceID = Tuple[int, Optional[str], int]

def _init_distributed_environment(
        parallel_config: ParallelConfig,
        rank: int,
        distributed_init_method: str,
        gpu_ids: List[int],
    ) -> None:
        print(f'deepspeed inference worker before::rank:{rank} CUDA_VISIBLE_DEVICES:{os.environ["CUDA_VISIBLE_DEVICES"]}',flush=True)
        if parallel_config.backend == "nccl":
            # Same as in Ray Train
            os.environ["NCCL_ASYNC_ERROR_HANDLING"] = "1"
            # All workers on a same node should share the same set of
            # visible GPUs. Otherwise they can't talk among themselves.
            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str(gid) for gid in gpu_ids)
            if "NCCL_SOCKET_IFNAME" not in os.environ:
                os.environ["NCCL_SOCKET_IFNAME"] = DEFAULT_NCCL_SOCKET_IFNAME

        os.environ["RANK"] = str(rank)
        os.environ["LOCAL_RANK"] = str(rank)
        os.environ["WORLD_SIZE"] = str(parallel_config.world_size)
        os.environ["LOCAL_WORLD_SIZE"] = str(parallel_config.world_size)        
        print(f'deepspeed inference worker after:rank:{rank} CUDA_VISIBLE_DEVICES:{os.environ["CUDA_VISIBLE_DEVICES"]}',flush=True)
        torch.cuda.set_device(rank)
        """Initialize the distributed environment."""
        torch.distributed.init_process_group(
            backend="nccl",
            world_size=parallel_config.world_size,
            rank=rank,
            init_method=distributed_init_method,            
        )
        # A small all_reduce for warmup.
        torch.distributed.all_reduce(torch.zeros(1).cuda())

        

class Worker:
    
    def __init__(
        self,        
        parallel_config: ParallelConfig,        
        rank: int,
        distributed_init_method: str,
    ) -> None:
        self.parallel_config = parallel_config        
        self.rank = rank
        self.distributed_init_method = distributed_init_method                
        self.model = None
        self.tokenizer = None
       

    def init_model(self,gpu_ids:List[int]):
        # Initialize the distributed environment.
        _init_distributed_environment(self.parallel_config, self.rank,
                                      self.distributed_init_method,gpu_ids)
        
        print(f"deepspeed inference worker:rank:{self.rank} load model {self.parallel_config.model_dir}",flush=True)
        tokenizer = AutoTokenizer.from_pretrained(self.parallel_config.model_dir,trust_remote_code=True)  
        model = AutoModelForCausalLM.from_pretrained(self.parallel_config.model_dir,trust_remote_code=True)       
        model = model.eval()
    
        ds_engine = deepspeed.init_inference(model,
                                mp_size=self.parallel_config.world_size,
                                dtype=torch.half,
                                replace_method="auto",
                                replace_with_kernel_inject=True)
        self.model = ds_engine.module
        self.tokenizer = tokenizer  
        print(f"deepspeed inference worker:rank:{self.rank} init successfully",flush=True)    

    def execute_model(self,ins:str, his:List[Tuple[str,str]]=[],  
        max_length:int=4096, 
        top_p:float=0.95,
        temperature:float=0.1,**kwargs):
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        tokens = self.tokenizer(ins, return_token_type_ids=False,return_tensors="pt").to(device)
        response = self.model.generate(
            input_ids=tokens["input_ids"],
            max_new_tokens=max_length,
            repetition_penalty=1.05,
            temperature=temperature,
            attention_mask=tokens.attention_mask        
        )
        answer = self.tokenizer.decode(response[0][tokens["input_ids"].shape[1]:], skip_special_tokens=True)
        return [(answer,"")]  
           

class DeepSpeedInference:
    def __init__(self,parallel_config: ParallelConfig ):    

        master_addr, master_port = get_address_and_port()        
        distributed_init_method = f"tcp://{master_addr}:{master_port}"  
        print(f"deepspeed inference: master_addr:{master_addr},master_port:{master_port}",flush=True)
        workers = []
        gpu_ids = ray.get_gpu_ids()
        gpu_ids_str = ",".join([str(gpu) for gpu in gpu_ids])
        
        for rank in range(parallel_config.world_size):    
            worker_cls = Worker  
            # deepspeed will use rank as the device id, and the 
            # ray will automatically set CUDA_VISIBLE_DEVICES for each worker according to the num_gpus
            # for example, suppose we have 0,1,2,4 gpus, and we have 4 workers, then the CUDA_VISIBLE_DEVICES 
            # for the last worker will be 3, and the deepspeed will use 3 as the device id, which is wrong because
            # he can only see one gpu. So we need to set CUDA_VISIBLE_DEVICES to 0,1,2,3 for each worker.
            runtime_env = {"env_vars": {
              "RAY_EXPERIMENTAL_NOSET_CUDA_VISIBLE_DEVICES":"true",
              "CUDA_VISIBLE_DEVICES":gpu_ids_str
            }}    
            worker_cls = ray.remote(
                        num_cpus=0,
                        num_gpus=0,
                        resources={f"node:{master_addr}": 1e-3},
                        runtime_env=runtime_env,                        
                    )(worker_cls).remote
            worker = worker_cls(parallel_config,rank,distributed_init_method)
            workers.append(worker)
        self.workers  = workers           
        ray.get([worker.init_model.remote(gpu_ids) for worker in self.workers])

    def stream_chat(self,tokenizer,ins:str, his:List[Tuple[str,str]]=[],  
        max_length:int=1024, 
        top_p:float=0.95,
        temperature:float=0.1,**kwargs):        
        output = self._run_workers("execute_model",ins,his,max_length,top_p,temperature)
        return output
              
    def _run_workers(
        self,
        method: str,
        *args,
        get_all_outputs: bool = False,
        **kwargs,
    ) -> Any:
        """Runs the given method on all workers."""
        all_outputs = []
        for worker in self.workers:
            executor = getattr(worker, method)
            executor = executor.remote
            output = executor(*args, **kwargs)
            all_outputs.append(output)

        all_outputs = ray.get(all_outputs)            
        
        if get_all_outputs:
            return all_outputs

        # Make sure all workers have the same results.
        output = all_outputs[0]
        for other_output in all_outputs[1:]:
            assert output == other_output
        return output    


  
    




from dataclasses import dataclass
from typing import TYPE_CHECKING,Dict, List, Optional, Union,Any,get_type_hints,Annotated,get_args
import typing
from .agent import Agent
from ray.util.client.common import ClientActorHandle, ClientObjectRef
import ray
import dataclasses
import copy
import inspect
import json
import pydantic


if TYPE_CHECKING:
    from .conversable_agent import ConversableAgent
    from byzerllm.utils.client import ByzerLLM,ByzerRetrieval

@dataclasses.dataclass
class ChatResponse:
      status: int
      output: str      
      code: str
      prompt: str
      variables: Dict[str,Any]=dataclasses.field(default_factory=dict)    
      raw_output: str = ""  

def get_agent_name(agent: Union[Agent,ClientActorHandle,str]) -> str:
    if isinstance(agent,Agent):
        agent_name = agent.name
    elif isinstance(agent,str):
        agent_name = agent
    else:    
        agent_name = ray.get(agent.get_name.remote())

    return agent_name    

def run_agent_func(agent: Union[Agent,"ConversableAgent",ClientActorHandle], func_name: str, *args, **kwargs):
    """Run a function of an agent."""
    if isinstance(agent,Agent):
        return getattr(agent, func_name)(*args, **kwargs)
    elif isinstance(agent,str):
        return ray.get(getattr(ray.get_actor(agent), func_name).remote(*args, **kwargs))    
    else:
        return ray.get(getattr(agent, func_name).remote(*args, **kwargs)) 
    
def count_messages_length(messages: List[Dict]) -> int:
    return sum([len(message["content"]) for message in messages])    
    
def copy_message(message: Dict) -> Dict:
    return copy.deepcopy(message)

def modify_message_metadata(message: Dict, **kwargs) -> Dict:
    message = copy_message(message)
    if "metadata" not in message:
        message["metadata"] = {}
    for key, value in kwargs.items():
        message["metadata"][key] = value
    return message

def modify_message_content(message: Dict, content:str) -> Dict:
    message = copy_message(message)
    message["content"] = content
    return message

def modify_last_message(messages: List[Dict], message:Dict) -> List[Dict]:
    messages = copy_message(messages)
    messages[-1] = message
    return messages

class Agents:
    @staticmethod
    def create_remote_agent(cls,name:str,llm,retrieval,*args, **kwargs)->ClientActorHandle:
        return ray.remote(name=name,max_concurrency=10)(cls).remote(
        name=name,llm=llm,retrieval=retrieval,*args, **kwargs)

    @staticmethod
    def create_remote_detached_agent(cls,name:str,llm,retrieval,*args, **kwargs)->ClientActorHandle:
        return ray.remote(name=name,max_concurrency=10,lifetime="detached"
                          )(cls).remote(
        name=name,llm=llm,retrieval=retrieval,*args, **kwargs)  

    @staticmethod
    def create_local_agent(cls,name:str,llm,retrieval,*args, **kwargs)->"ConversableAgent":
        return cls(name=name,llm=llm,retrieval=retrieval,*args, **kwargs)

    @staticmethod
    def create_local_group(group_name:str,agents: List[Agent],llm,retrieval,*args, **kwargs) -> List["ConversableAgent"]:
        from .groupchat import GroupChat
        from .groupchat import GroupChatManager

        if any([not isinstance(agent,Agent) for agent in agents]):
            raise ValueError("agents must be a list of Agent objects")
                
        group_parameters = ["messages","max_round","admin_name","func_call_filter"]
        group_parameters_dict = {}
        for parameter in group_parameters:
            if parameter in kwargs:
                group_parameters_dict[parameter] = kwargs[parameter]
                del kwargs[parameter]

        groupchat = GroupChat(agents=agents, **group_parameters_dict)
        group_chat_manager =Agents.create_local_agent(GroupChatManager,name=group_name,
                                                       llm=llm,retrieval=retrieval,
                                                       groupchat=groupchat,*args, **kwargs)
        return group_chat_manager
    
    @staticmethod
    def create_remote_group(group_name:str,agents: List[Union[Agent,ClientActorHandle,str]],llm,retrieval,*args, **kwargs) -> List[Union[Agent,ClientActorHandle,str]]:
        from .groupchat import GroupChat
        from .groupchat import GroupChatManager
                
        group_parameters = ["messages","max_round","admin_name","func_call_filter"]
        group_parameters_dict = {}
        for parameter in group_parameters:
            if parameter in kwargs:
                group_parameters_dict[parameter] = kwargs[parameter]
                del kwargs[parameter]

        groupchat = GroupChat(agents=agents, **group_parameters_dict)
        group_chat_manager =Agents.create_remote_agent(GroupChatManager,name=group_name,
                                                       llm=llm,retrieval=retrieval,
                                                       groupchat=groupchat,*args, **kwargs)
        return group_chat_manager   



        


import os
from os.path import expanduser
import urllib.request
import tarfile
from loguru import logger
import json
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress
from rich import print as rprint

from byzerllm.apps.byzer_storage.env import get_latest_byzer_retrieval_lib
from byzerllm.apps.byzer_storage import env
from modelscope.hub.snapshot_download import snapshot_download
from byzerllm.utils.client import InferBackend
import huggingface_hub
import byzerllm
from pydantic import BaseModel
from huggingface_hub import snapshot_download as hf_snapshot_download
from typing import List, Optional, Tuple

try:
    import torch
except ImportError:
    torch = None


class StorageLocation(BaseModel):
    version: str
    cluster: str
    base_dir: str
    libs_dir: str
    data_dir: str


console = Console()


def check_dependencies():
    import importlib
    from packaging import version

    dependencies = {"vllm": "0.4.3", "torch": "2.3.0", "flash_attn": "2.5.9.post1"}

    for package, required_version in dependencies.items():
        try:
            module = importlib.import_module(package)
            current_version = module.__version__
            if version.parse(current_version) != version.parse(required_version):
                console.print(
                    f"[yellow]Warning: {package} version mismatch. Required: {required_version}, Found: {current_version}[/yellow]"
                )
        except ImportError:
            console.print(f"[yellow]Warning: {package} is not installed.[/yellow]")


class StorageSubCommand:

    @staticmethod
    def download_model(
        model_id: str,
        cache_dir: str,
        local_files_only: bool = False,
        use_huggingface: bool = False,
    ) -> Tuple[Optional[str], List[str]]:
        error_summary = []
        console.print(f"[bold blue]Downloading model {model_id}...")
        try:
            if use_huggingface:
                model_path = hf_snapshot_download(
                    repo_id=model_id,
                    cache_dir=cache_dir,
                    local_files_only=local_files_only,
                )
            else:
                model_path = snapshot_download(
                    model_id=model_id,
                    cache_dir=cache_dir,
                    local_files_only=local_files_only,
                )
            console.print(f"[green]✓[/green] Model downloaded: {model_path}")
            return model_path, error_summary
        except Exception as e:
            error_msg = f"Failed to download model: {str(e)}"
            console.print(f"[red]✗[/red] {error_msg}")
            console.print(
                f"[yellow]![/yellow] Please manually download the model '{model_id}'"
            )
            console.print(
                f"[yellow]![/yellow] and place it in the directory: {cache_dir}"
            )
            error_summary.append(error_msg)
            return None, error_summary

    @staticmethod
    def emb_start(args):
        error_summary = []
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval

        home = expanduser("~")
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        base_model_dir = os.path.join(base_dir, "storage", "models")
        bge_model = os.path.join(base_model_dir, "AI-ModelScope", "bge-large-zh")

        console.print("[bold blue]Starting embedding model...")

        if not os.path.exists(bge_model):
            bge_model, download_errors = StorageSubCommand.download_model(
                "AI-ModelScope/bge-large-zh",
                base_model_dir,
                local_files_only=huggingface_hub.constants.HF_HUB_OFFLINE,
            )
            error_summary.extend(download_errors)
            if not bge_model:
                error_summary.append("Failed to download embedding model")
                return error_summary

        try:
            byzerllm.connect_cluster(address=args.ray_address)
            llm = byzerllm.ByzerLLM()

            if llm.is_model_exist("emb"):
                console.print("[yellow]![/yellow] Embedding model already deployed.")
                return error_summary

            console.print("[bold blue]Deploying embedding model...")
            llm.setup_num_workers(1).setup_infer_backend(InferBackend.Transformers)

            if torch and torch.cuda.is_available():
                llm.setup_gpus_per_worker(0.1)
            else:
                llm.setup_gpus_per_worker(0)

            llm.setup_cpus_per_worker(0.01).setup_worker_concurrency(20)
            llm.deploy(
                model_path=bge_model,
                pretrained_model_type="custom/bge",
                udf_name="emb",
                infer_params={},
            )
            console.print("[green]✓[/green] Embedding model deployed successfully")
        except Exception as e:
            error_msg = f"Failed to deploy embedding model: {str(e)}"
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)

        return error_summary

    @staticmethod
    def emb_stop(args):
        error_summary = []
        import byzerllm

        console.print("[bold blue]Stopping embedding model...")
        try:
            byzerllm.connect_cluster(address=args.ray_address)
            llm = byzerllm.ByzerLLM()

            llm.undeploy("emb")
            console.print("[green]✓[/green] Embedding model stopped successfully")
        except Exception as e:
            error_msg = f"Failed to stop embedding model: {str(e)}"
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)

        return error_summary

    @staticmethod
    def model_memory_start(args):
        error_summary = []
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval

        home = expanduser("~")
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        base_model_dir = os.path.join(base_dir, "storage", "models")
        llama_model = os.path.join(
            base_model_dir, "meta-llama", "Meta-Llama-3-8B-Instruct-GPTQ"
        )

        console.print("[bold blue]Starting long-term memory model...")

        if not os.path.exists(llama_model):
            llama_model, download_errors = StorageSubCommand.download_model(
                "meta-llama/Meta-Llama-3-8B-Instruct-GPTQ", base_model_dir
            )
            error_summary.extend(download_errors)
            if not llama_model:
                error_summary.append("Failed to download long-term memory model")
                return error_summary

        if not torch or not torch.cuda.is_available():
            error_msg = "GPU not available. Long-term memory model requires GPU."
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)
            return error_summary

        try:
            byzerllm.connect_cluster(address=args.ray_address)
            llm = byzerllm.ByzerLLM()

            if llm.is_model_exist("long_memory"):
                console.print(
                    "[yellow]![/yellow] Long-term memory model already deployed."
                )
                return error_summary

            console.print("[bold blue]Checking dependencies...")
            check_dependencies()

            console.print("[bold blue]Deploying long-term memory model...")
            llm.setup_gpus_per_worker(1).setup_cpus_per_worker(0.001).setup_num_workers(
                1
            )
            llm.setup_infer_backend(InferBackend.VLLM)
            llm.deploy(
                model_path=llama_model,
                pretrained_model_type="custom/auto",
                udf_name="long_memory",
                infer_params={
                    "backend.max_model_len": 8000,
                    "backend.enable_lora": True,
                },
            )
            console.print(
                "[green]✓[/green] Long-term memory model deployed successfully"
            )
        except Exception as e:
            error_msg = f"Failed to deploy long-term memory model: {str(e)}"
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)

        return error_summary

    @staticmethod
    def model_memory_stop(args):
        error_summary = []
        import byzerllm

        console.print("[bold blue]Stopping long-term memory model...")
        try:
            byzerllm.connect_cluster(address=args.ray_address)
            llm = byzerllm.ByzerLLM()

            llm.undeploy("long_memory")
            console.print(
                "[green]✓[/green] Long-term memory model stopped successfully"
            )
        except Exception as e:
            error_msg = f"Failed to stop long-term memory model: {str(e)}"
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)

        return error_summary

    @staticmethod
    def install(args):
        error_summary = []
        version = args.version
        home = expanduser("~")
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")
        libs_dir = os.path.join(
            base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}"
        )
        if os.path.exists(libs_dir):
            logger.info(f"Byzer Storage version {version} already installed.")
            return error_summary

        try:
            if not os.path.exists(base_dir):
                os.makedirs(base_dir, exist_ok=True)

            env_info = env.detect_env()

            logger.info("Current environment:")
            logger.info(env_info)

            if env_info.java_version == "" or int(env_info.java_version) < 21:
                logger.info("JDK 21 not found, downloading and installing JDK 21...")
                try:
                    env.download_and_install_jdk21(env_info, base_dir)
                except Exception as e:
                    error_msg = f"Error downloading and installing JDK 21: {str(e)}. You may need to install JDK 21 manually."
                    logger.error(error_msg)
                    error_summary.append(error_msg)

            download_url = f"https://download.byzer.org/byzer-retrieval/byzer-retrieval-lib-{version}.tar.gz"
            libs_dir = os.path.join(base_dir, "storage", "libs")

            os.makedirs(libs_dir, exist_ok=True)
            download_path = os.path.join(
                libs_dir, f"byzer-retrieval-lib-{version}.tar.gz"
            )
            if os.path.exists(download_path):
                logger.info(f"Byzer Storage version {version} already downloaded.")
            else:

                def download_with_progressbar(url, filename):
                    def progress(count, block_size, total_size):
                        percent = int(count * block_size * 100 / total_size)
                        print(f"\rDownload progress: {percent}%", end="")

                    urllib.request.urlretrieve(url, filename, reporthook=progress)

                logger.info(f"Download Byzer Storage version {version}: {download_url}")
                download_with_progressbar(download_url, download_path)

                with tarfile.open(download_path, "r:gz") as tar:
                    tar.extractall(path=libs_dir)

        except Exception as e:
            error_msg = f"Failed to install Byzer Storage: {str(e)}"
            logger.error(error_msg)
            error_summary.append(error_msg)

        return error_summary

    @staticmethod
    def collection(args):
        error_summary = []
        try:
            from byzerllm.apps.llama_index.collection_manager import (
                CollectionManager,
                CollectionItem,
            )

            home = expanduser("~")
            base_dir = args.base_dir or os.path.join(home, ".auto-coder")
            collection_manager = CollectionManager(base_dir)
            if args.name:
                collection = CollectionItem(
                    name=args.name, description=args.description
                )
                collection_manager.add_collection(collection)
                console.print(
                    f"[green]✓[/green] Collection {args.name} added successfully."
                )
            else:
                error_msg = "Please provide collection name."
                console.print(f"[red]✗[/red] {error_msg}")
                error_summary.append(error_msg)
        except Exception as e:
            error_msg = f"Failed to add collection: {str(e)}"
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)

        return error_summary

    @staticmethod
    def get_store_location(args):
        version = args.version
        cluster = args.cluster
        home = expanduser("~")
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")

        libs_dir = os.path.join(
            base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}"
        )
        data_dir = os.path.join(base_dir, "storage", "data")
        return StorageLocation(
            version=version,
            cluster=cluster,
            base_dir=base_dir,
            libs_dir=libs_dir,
            data_dir=data_dir,
        )

    @staticmethod
    def connect_cluster(args):
        error_summary = []
        store_location = StorageSubCommand.get_store_location(args)

        console.print("[bold green]Starting Byzer Storage...[/bold green]")

        try:
            if not os.path.exists(
                os.path.join(store_location.data_dir, store_location.cluster)
            ):
                os.makedirs(store_location.data_dir, exist_ok=True)
                rprint("[green]✓[/green] Created data directory")

            if not os.path.exists(store_location.libs_dir):
                rprint("[bold blue]Installing Byzer Storage...[/bold blue]")
                install_errors = StorageSubCommand.install(args)
                error_summary.extend(install_errors)
                if not install_errors:
                    rprint("[green]✓[/green] Installed Byzer Storage")

            code_search_path = [store_location.libs_dir]

            console.print("[bold blue]Connecting to cluster...[/bold blue]")
            env_vars = byzerllm.connect_cluster(
                address=args.ray_address, code_search_path=code_search_path
            )
            rprint("[green]✓[/green] Connected to cluster")
        except Exception as e:
            error_msg = f"Failed to connect to cluster: {str(e)}"
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)
            return None, None, error_summary

        return env_vars, store_location, error_summary

    @staticmethod
    def start(args):
        error_summary = []
        from byzerllm.utils.retrieval import ByzerRetrieval

        env_vars, store_location, connect_errors = StorageSubCommand.connect_cluster(
            args
        )
        error_summary.extend(connect_errors)

        if env_vars is None or store_location is None:
            return error_summary

        try:
            console.print("[bold blue]Launching gateway...[/bold blue]")
            retrieval = ByzerRetrieval()
            retrieval.launch_gateway()
            rprint("[green]✓[/green] Launched gateway")

            if retrieval.is_cluster_exists(name=store_location.cluster):
                error_msg = f"Cluster {store_location.cluster} already exists. Please stop it first."
                console.print(Panel(f"[yellow]{error_msg}[/yellow]"))
                error_summary.append(error_msg)
                return error_summary

            base_model_dir = os.path.join(store_location.base_dir, "storage", "models")
            os.makedirs(base_model_dir, exist_ok=True)

            console.print("[bold blue]Checking GPU availability...[/bold blue]")
            has_gpu = torch and torch.cuda.is_available()
            if has_gpu:
                rprint("[green]✓[/green] GPU detected")
            else:
                rprint("[yellow]![/yellow] No GPU detected, using CPU")

            if args.enable_emb:
                emb_errors = StorageSubCommand.emb_start(args)
                error_summary.extend(emb_errors)

            if args.enable_model_memory and has_gpu:
                memory_errors = StorageSubCommand.model_memory_start(args)
                error_summary.extend(memory_errors)

            cluster_json = os.path.join(
                store_location.base_dir,
                "storage",
                "data",
                f"{store_location.cluster}.json",
            )
            if os.path.exists(cluster_json):
                console.print("[bold blue]Restoring Byzer Storage...[/bold blue]")
                restore_errors = StorageSubCommand.restore(args)
                error_summary.extend(restore_errors)
                if not restore_errors:
                    console.print(
                        Panel(
                            "[green]Byzer Storage restored and started successfully[/green]"
                        )
                    )
                return error_summary

            console.print("[bold blue]Starting cluster...[/bold blue]")
            try:
                builder = retrieval.cluster_builder()
                builder.set_name(store_location.cluster).set_location(
                    store_location.data_dir
                ).set_num_nodes(args.num_nodes).set_node_cpu(
                    args.node_cpus
                ).set_node_memory(
                    f"{args.node_memory}g"
                )
                builder.set_java_home(env_vars["JAVA_HOME"]).set_path(
                    env_vars["PATH"]
                ).set_enable_zgc()
                builder.start_cluster()

                with open(
                    os.path.join(
                        store_location.base_dir,
                        "storage",
                        "data",
                        f"{store_location.cluster}.json",
                    ),
                    "w",
                ) as f:
                    f.write(
                        json.dumps(
                            retrieval.cluster_info(store_location.cluster),
                            ensure_ascii=False,
                        )
                    )

                console.print(
                    Panel("[green]Byzer Storage started successfully[/green]")
                )
            except Exception as e:
                error_msg = f"Failed to start Byzer Storage cluster: {str(e)}"
                console.print(f"[red]✗[/red] {error_msg}")
                error_summary.append(error_msg)
        except Exception as e:
            error_msg = f"An unexpected error occurred: {str(e)}"
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)

        return error_summary

    @staticmethod
    def stop(args):
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval

        error_summary = []
        store_location = StorageSubCommand.get_store_location(args)
        console.print("[bold red]Stopping Byzer Storage...")
        libs_dir = os.path.join(
            store_location.base_dir,
            "storage",
            "libs",
            f"byzer-retrieval-lib-{store_location.version}",
        )
        cluster_json = os.path.join(
            store_location.base_dir, "storage", "data", f"{store_location.cluster}.json"
        )

        if not os.path.exists(cluster_json) or not os.path.exists(libs_dir):
            error_msg = "No instance found. Please check if Byzer Storage is properly installed."
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)
            return error_summary

        env_vars, _, connect_errors = StorageSubCommand.connect_cluster(args)
        error_summary.extend(connect_errors)

        if env_vars is None:
            return error_summary

        console.print("[bold blue]Launching gateway...")
        try:
            retrieval = ByzerRetrieval()
            retrieval.launch_gateway()
            rprint("[green]✓[/green] Gateway launched")
        except Exception as e:
            error_msg = f"Failed to launch gateway: {str(e)}"
            rprint(f"[red]✗[/red] {error_msg}")
            error_summary.append(
                f"{error_msg} Please check if Byzer Retrieval is properly installed."
            )

        console.print(f"[bold blue]Shutting down cluster {store_location.cluster}...")
        try:
            retrieval.shutdown_cluster(cluster_name=store_location.cluster)
            rprint(f"[green]✓[/green] Cluster {store_location.cluster} shut down")
        except Exception as e:
            error_msg = (
                f"Failed to shut down cluster {store_location.cluster}: {str(e)}"
            )
            rprint(f"[red]✗[/red] {error_msg}")
            error_summary.append(f"{error_msg} You may need to manually stop it.")

        emb_stop_errors = StorageSubCommand.emb_stop(args)
        error_summary.extend(emb_stop_errors)

        model_memory_stop_errors = StorageSubCommand.model_memory_stop(args)
        error_summary.extend(model_memory_stop_errors)

        if error_summary:
            console.print(
                Panel("[yellow]Byzer Storage stopped with some issues[/yellow]")
            )
            console.print(
                "[bold red]The following steps need manual attention:[/bold red]"
            )
            for error in error_summary:
                console.print(f"- {error}")
        else:
            console.print(Panel("[green]Byzer Storage stopped successfully[/green]"))

        return error_summary

    @staticmethod
    def export(args):
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval

        error_summary = []
        version = args.version
        cluster = args.cluster
        home = expanduser("~")
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")

        console.print("[bold blue]Exporting Byzer Storage...")
        libs_dir = os.path.join(
            base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}"
        )
        cluster_json = os.path.join(base_dir, "storage", "data", f"{cluster}.json")

        if not os.path.exists(cluster_json) or not os.path.exists(libs_dir):
            error_msg = "No instance found. Please check if Byzer Storage is properly installed."
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)
            return error_summary

        code_search_path = [libs_dir]

        console.print("[bold blue]Connecting to cluster...")
        try:
            byzerllm.connect_cluster(
                address=args.ray_address, code_search_path=code_search_path
            )
            rprint("[green]✓[/green] Connected to cluster")
        except Exception as e:
            error_msg = f"Failed to connect to cluster: {str(e)}"
            rprint(f"[red]✗[/red] {error_msg}")
            error_summary.append(
                f"{error_msg} Please check your network connection and Ray setup."
            )
            return error_summary

        try:
            retrieval = ByzerRetrieval()
            retrieval.launch_gateway()
            rprint("[green]✓[/green] Gateway launched")
        except Exception as e:
            error_msg = f"Failed to launch gateway: {str(e)}"
            rprint(f"[red]✗[/red] {error_msg}")
            error_summary.append(
                f"{error_msg} Please check if Byzer Retrieval is properly installed."
            )
            return error_summary

        console.print(f"[bold blue]Exporting cluster {cluster} information...")
        try:
            cluster_info = retrieval.cluster_info(cluster)
            with open(cluster_json, "w") as f:
                json.dump(cluster_info, f, ensure_ascii=False, indent=2)
            rprint(
                f"[green]✓[/green] Cluster {cluster} information exported to {cluster_json}"
            )
        except Exception as e:
            error_msg = f"Failed to export cluster {cluster} information: {str(e)}"
            rprint(f"[red]✗[/red] {error_msg}")
            error_summary.append(
                f"{error_msg} You may need to check the cluster status."
            )

        if error_summary:
            console.print(
                Panel("[yellow]Byzer Storage exported with some issues[/yellow]")
            )
            console.print(
                "[bold red]The following steps need manual attention:[/bold red]"
            )
            for error in error_summary:
                console.print(f"- {error}")
        else:
            console.print(Panel("[green]Byzer Storage exported successfully[/green]"))

        return error_summary

    @staticmethod
    def restore(args):
        import byzerllm
        from byzerllm.utils.retrieval import ByzerRetrieval

        error_summary = []
        version = args.version
        cluster = args.cluster
        home = expanduser("~")
        base_dir = args.base_dir or os.path.join(home, ".auto-coder")

        console.print("[bold blue]Restoring Byzer Storage...")
        libs_dir = os.path.join(
            base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}"
        )
        cluster_json = os.path.join(base_dir, "storage", "data", f"{cluster}.json")

        if not os.path.exists(cluster_json) or not os.path.exists(libs_dir):
            error_msg = "No instance found. Please check if Byzer Storage is properly installed."
            console.print(f"[red]✗[/red] {error_msg}")
            error_summary.append(error_msg)
            return error_summary

        code_search_path = [libs_dir]

        console.print("[bold blue]Connecting to cluster...")
        try:
            byzerllm.connect_cluster(
                address=args.ray_address, code_search_path=code_search_path
            )
            rprint("[green]✓[/green] Connected to cluster")
        except Exception as e:
            error_msg = f"Failed to connect to cluster: {str(e)}"
            rprint(f"[red]✗[/red] {error_msg}")
            error_summary.append(
                f"{error_msg} Please check your network connection and Ray setup."
            )
            return error_summary

        console.print("[bold blue]Launching gateway...")
        try:
            retrieval = ByzerRetrieval()
            retrieval.launch_gateway()
            rprint("[green]✓[/green] Gateway launched")
        except Exception as e:
            error_msg = f"Failed to launch gateway: {str(e)}"
            rprint(f"[red]✗[/red] {error_msg}")
            error_summary.append(
                f"{error_msg} Please check if Byzer Retrieval is properly installed."
            )
            return error_summary

        console.print(f"[bold blue]Restoring cluster {cluster}...")
        try:
            if not retrieval.is_cluster_exists(cluster):
                with open(cluster_json, "r") as f:
                    cluster_info = json.load(f)
                retrieval.restore_from_cluster_info(cluster_info)
                rprint(f"[green]✓[/green] Cluster {cluster} restored successfully")
            else:
                warning_msg = (
                    f"Cluster {cluster} already exists. No restoration needed."
                )
                rprint(f"[yellow]![/yellow] {warning_msg}")
                error_summary.append(warning_msg)
        except Exception as e:
            error_msg = f"Failed to restore cluster {cluster}: {str(e)}"
            rprint(f"[red]✗[/red] {error_msg}")
            error_summary.append(
                f"{error_msg} You may need to check the cluster status and configuration."
            )

        if error_summary:
            console.print(
                Panel("[yellow]Byzer Storage restored with some issues[/yellow]")
            )
            console.print(
                "[bold red]The following steps need manual attention:[/bold red]"
            )
            for error in error_summary:
                console.print(f"- {error}")
        else:
            console.print(Panel("[green]Byzer Storage restored successfully[/green]"))

        return error_summary



from langchain.embeddings.base import Embeddings
from typing import Any, List, Mapping, Optional,Tuple
from langchain.callbacks.manager import CallbackManagerForLLMRun
from langchain.llms.base import LLM
import requests
import json

from . import ClientParams


class ByzerLLMClient:
    
    def __init__(self,url:str='http://127.0.0.1:9003/model/predict',params:ClientParams=ClientParams()) -> None:
        self.url = url
        self.client_params = params        

    def request(self, sql:str,json_data:str)->str:         
        data = {
            'sessionPerUser': 'true',
            'sessionPerRequest': 'true',
            'owner': self.client_params.owner,
            'dataType': 'string',
            'sql': sql,
            'data': json_data
        }
        response = requests.post(self.url, data=data)
        if response.status_code != 200:
            raise Exception(f"{self.url} status:{response.status_code} content: {response.text} request: json/{json.dumps(data,ensure_ascii=False)}")
        return response.text

    def emb(self,s:str)-> List[float]:
        json_data = json.dumps([
            {"instruction":s,"embedding":True}
        ],ensure_ascii=False)
        response = self.request(f'''
        select {self.client_params.llm_embedding_func}(array(feature)) as value
        ''',json_data)    
        t = json.loads(response)
        t2 = json.loads(t[0]["value"][0])
        return t2[0]["predict"]

    def chat(self,s:str,history:List[Tuple[str,str]],extra_query={})->str:
        newhis = [{"query":item[0],"response":item[1]} for item in history]
        json_data = json.dumps([
            {"instruction":s,"history":newhis,**extra_query}
        ],ensure_ascii=False)
        response = self.request(f'''
        select {self.client_params.llm_chat_func}(array(feature)) as value
        ''',json_data)    
        t = json.loads(response)
        t2 = json.loads(t[0]["value"][0])
        return t2[0]["predict"]


class LocalEmbeddings(Embeddings):
    def __init__(self,client:ByzerLLMClient,prompt_prefix=None):
        self.client = client
        self.prompt_prefix = prompt_prefix
                
        
    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        embeddings = []        
        for text in texts:
            if self.prompt_prefix:
                text = self.prompt_prefix + text
            embedding = self.client.emb(text)
            embeddings.append(embedding)        
        return embeddings

    def embed_query(self, text: str) -> List[float]: 
        if self.prompt_prefix:
            text = self.prompt_prefix + text   
        embedding = self.client.emb(text)
        return embedding


class Chatglm6bLLM(LLM):
    
    def __init__(self,client:ByzerLLMClient):
        self.client = client
        
    @property
    def _llm_type(self) -> str:
        return "chatglm6b"
    
    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
    ) -> str:
        if stop is not None:
            raise ValueError("stop kwargs are not permitted.")
        return self.client.chat(prompt,[])
    
    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {"n": self.n}


import io

from langchain.docstore.document import Document
from abc import ABC, abstractmethod
from typing import List, Tuple,Dict,Any
import json
import csv


class DocRetrieveStrategy(ABC):
    @abstractmethod
    def retrieve(self, docs: List[Tuple[Document, float]], k: int) -> List[Tuple[Document, float]]:
        pass


class FullDocRetrieveStrategy(DocRetrieveStrategy):
    def retrieve(self, docs: List[Tuple[Document, float]], k: int) -> List[Tuple[Document, float]]:
        if not docs:            
            return []

        doc_hits = {}
        for doc in docs:
            source = doc[0].metadata['source']            
            doc,score,hits = doc_hits.get(source, (doc[0], doc[1], 0))
            doc_hits[source] = (doc,score,hits+1)

        # Sort by hits descending and then by score ascending
        sorted_docs = sorted(doc_hits.values(), key=lambda x: (x[2], -1 * x[1]), reverse=True)
        doc,score,hits = sorted_docs[0]
        doc.page_content = doc.metadata['page_content']
        return [(doc, score)]


class DocRetrieveStrategyFactory(DocRetrieveStrategy):
    def __init__(self, strategy: str) -> None:
        self.strategy = strategy

    def retrieve(self, docs: List[Tuple[Document, float]], k: int) -> List[Tuple[Document, float]]:
        if self.strategy == "full_doc":
            print("Using full_doc strategy",flush=True)
            return FullDocRetrieveStrategy().retrieve(docs, k)
        else:
            return docs[0:k]


class DocCombineFormat(ABC):
    @abstractmethod
    def combine(self, docs: List[Tuple[Document, float]], k: int) -> Tuple[str, List[Dict[Any, Any]]]:
        pass


class FullDocCombineFormatList(DocCombineFormat):
    def __init__(self, input: Dict[str,Any]) -> None:
        self.params = input

    def combine(self, docs: List[Tuple[Document, float]], k: int) -> Tuple[str, List[Dict[Any, Any]]]:
        if not docs :
            return None

        temp_docs = []
        temp_metas = []
        for index, doc in enumerate(docs[0:k]):
            temp_docs.append(f'{index}. {doc[0].page_content}')
            if "metadata" in doc[0]:
                temp_metas.append(doc[0].metadata)
        return ("\n".join(temp_docs), temp_metas)


class FullDocCombineFormatDefault(DocCombineFormat):
    def __init__(self, input: Dict[str,Any]) -> None:
        self.params = input

    def combine(self, docs: List[Tuple[Document, float]], k: int) -> Tuple[str, List[Dict[Any, Any]]]:
        if docs is None or len(docs) == 0:
            return None

        temp_docs = []
        temp_metas = []
        for index, doc in enumerate(docs[0:k]):
            temp_docs.append(f'{doc[0].page_content}')
            if "metadata" in doc[0]:
                temp_metas.append(doc[0].metadata)
        return ("\n".join(temp_docs), temp_metas)

class JsonCombineFormat(DocCombineFormat):
    def __init__(self, input: Dict[str,Any]) -> None:
        self.params = input

    def combine(self, docs: List[Tuple[Document, float]], k: int) -> Tuple[str, List[Dict[Any, Any]]]:
        if docs is None or len(docs) == 0:
            return None

        temp_docs = []
        temp_metas = []
        for index, doc in enumerate(docs[0:k]):
            temp_docs.append({"body":doc[0].page_content})
            if "metadata" in doc[0]:
                temp_metas.append(doc[0].metadata)
        return (json.dumps(temp_docs,ensure_ascii=False), temp_metas)


class CsvCombineFormat(DocCombineFormat):
    def __init__(self, input: Dict[str,Any]) -> None:
        self.params = input

    def combine(self, docs: List[Tuple[Document, float]], k: int) -> Tuple[str, List[Dict[Any, Any]]]:
        if docs is None or len(docs) == 0:
            return None
                
        with io.StringIO() as csv_buffer:
            csv_writer = csv.DictWriter(csv_buffer, fieldnames=['body'])
            csv_writer.writeheader()
            temp_metas = []
            for index, doc in enumerate(docs[0:k]):
                csv_writer.writerow({"body": doc[0].page_content})
                if "metadata" in doc[0]:
                    temp_metas.append(doc[0].metadata)

            value = csv_buffer.getvalue()        
        return (value, temp_metas)


class FullDocCombineFormatFactory(DocCombineFormat):
    def __init__(self, input: Dict[str,Any]) -> None:
        self.params = input
        self.format = input.get("format","")       

    def combine(self, docs: List[Tuple[Document, float]], k: int) -> Tuple[str, List[Dict[Any, Any]]]:
        if self.format == 'list':
            return FullDocCombineFormatList(self.params).combine(docs, k)
        elif self.format == 'json':
            return JsonCombineFormat(self.params).combine(docs, k)
        elif self.format == 'csv':
            return CsvCombineFormat(self.params).combine(docs, k)
        else:
            return FullDocCombineFormatDefault(self.params).combine(docs, k)



import argparse
from byzerllm.byzerllm_command import main as byzerllm_main

MODEL_INFER_PARAMS_MAP = {
    "gpt-3.5-turbo-0125": "saas.api_key=${MODEL_TOKEN} saas.model=gpt-3.5-turbo-0125",
    "text-embedding-3-small": "saas.api_key=${MODEL_TOKEN} saas.model=text-embedding-3-small",
    "deepseek-chat": 'saas.base_url="https://api.deepseek.com/v1" saas.api_key=${MODEL_TOKEN} saas.model=deepseek-chat',
    "deepseek-coder": 'saas.base_url="https://api.deepseek.com/v1" saas.api_key=${MODEL_TOKEN} saas.model=deepseek-coder',
    "moonshot-v1-32k": 'saas.api_key=${MODEL_TOKEN} saas.base_url="https://api.moonshot.cn/v1" saas.model=moonshot-v1-32k',
    "qwen1.5-32b-chat": "saas.api_key=${MODEL_TOKEN} saas.model=qwen1.5-32b-chat",
    "qwen-max": "saas.api_key=${MODEL_TOKEN} saas.model=qwen-max",
    "alibaba/Qwen1.5-110B-Chat": 'saas.base_url="https://api.siliconflow.cn/v1" saas.api_key=${MODEL_TOKEN} saas.model=alibaba/Qwen1.5-110B-Chat',
    "deepseek-ai/deepseek-v2-chat": 'saas.base_url="https://api.siliconflow.cn/v1" saas.api_key=${MODEL_TOKEN} saas.model=deepseek-ai/deepseek-v2-chat',
    "alibaba/Qwen2-72B-Instruct": 'saas.base_url="https://api.siliconflow.cn/v1" saas.api_key=${MODEL_TOKEN} saas.model=alibaba/Qwen2-72B-Instruct',
    "qwen-vl-chat-v1": "saas.api_key=${MODEL_TOKEN} saas.model=qwen-vl-chat-v1",
    "qwen-vl-max": "saas.api_key=${MODEL_TOKEN} saas.model=qwen-vl-max",
    "yi-vision": "saas.api_key=${MODEL_TOKEN} saas.base_url=https://api.lingyiwanwu.com/v1 saas.model=yi-vision",
    "gpt4o": "saas.api_key=${MODEL_TOKEN} saas.model=gpt-4o",
    "sonnet3.5": "saas.api_key=${MODEL_TOKEN} saas.model=claude-3-5-sonnet-20240620",
}

import re

MODEL_PRETRAINED_TYPE_MAP = {
    "gpt-3.5-turbo-0125": "saas/openai",
    "text-embedding-3-small": "saas/openai",
    "deepseek-chat": "saas/openai",
    "deepseek-coder": "saas/openai",
    "moonshot-v1-32k": "saas/official_openai",
    "qwen1.5-32b-chat": "saas/qianwen",
    "qwen-max": "saas/qianwen",
    "alibaba/Qwen1.5-110B-Chat": "saas/openai",
    "deepseek-ai/deepseek-v2-chat": "saas/openai",
    "alibaba/Qwen2-72B-Instruct": "saas/openai",
    "qwen-vl-chat-v1": "saas/qianwen_vl",
    "qwen-vl-max": "saas/qianwen_vl",
    "yi-vision": "saas/openai",
    "gpt4o": "saas/openai",
    "sonnet3.5": "saas/claude",
}


def model_to_instance(model_name):
    return model_name


def main():
    parser = argparse.ArgumentParser(description="Easy ByzerLLM command line interface")
    subparsers = parser.add_subparsers(dest="command", required=True)

    deploy_parser = subparsers.add_parser("deploy", help="Deploy a model")
    deploy_parser.add_argument("model", type=str, help="Model name to deploy")
    deploy_parser.add_argument(
        "--token", type=str, required=True, help="The model token"
    )
    deploy_parser.add_argument(
        "--ray_address", default="auto", help="Ray cluster address to connect to"
    )
    deploy_parser.add_argument(
        "--infer_params", type=str, default="", help="Infer params for the model"
    )
    deploy_parser.add_argument(
        "--base_url", type=str, default="", help="base url"
    )
    deploy_parser.add_argument(
        "--alias", type=str, default="", help="Alias name for the deployed model"
    )

    undeploy_parser = subparsers.add_parser("undeploy", help="Undeploy a model")
    undeploy_parser.add_argument("model", type=str, help="Model name to undeploy")
    undeploy_parser.add_argument(
        "--ray_address", default="auto", help="Ray cluster address to connect to"
    )
    undeploy_parser.add_argument(
        "--force", action="store_true", help="Force undeploy the model"
    )

    chat_parser = subparsers.add_parser("chat", help="Chat with a deployed model")
    chat_parser.add_argument("model", type=str, help="Model name to chat with")
    chat_parser.add_argument("query", type=str, help="User query")
    chat_parser.add_argument(
        "--ray_address", default="auto", help="Ray cluster address to connect to"
    )

    args = parser.parse_args()
    instance_name = model_to_instance(args.model)

    if args.command == "deploy":
        instance_name = args.alias or model_to_instance(args.model)
        infer_params = args.infer_params or MODEL_INFER_PARAMS_MAP.get(args.model, "")
        
        if args.base_url:
            if "saas.base_url" in infer_params:
                infer_params = re.sub(r"saas.base_url=[^ ]+", f"saas.base_url={args.base_url}", infer_params)
            else:
                infer_params += f" saas.base_url={args.base_url}"

        pretrained_model_type = MODEL_PRETRAINED_TYPE_MAP.get(args.model, "")
        if not pretrained_model_type:
            raise ValueError(f"Pretrained model type not found for model {args.model}")

        infer_params = infer_params.replace("${MODEL_TOKEN}", args.token)

        byzerllm_main(
            [
                "deploy",
                "--model",
                instance_name,
                "--ray_address",
                args.ray_address,
                "--infer_params",
                infer_params,
                "--pretrained_model_type",
                pretrained_model_type,
                "--cpus_per_worker",
                "0.001",
                "--gpus_per_worker",
                "0",
                "--num_workers",
                "1",
                "--worker_concurrency",
                "1000",
            ]
        )
    elif args.command == "undeploy":
        target_args = [
            "undeploy",
            "--model",
            instance_name,
            "--ray_address",
            args.ray_address,
        ]
        if args.force:
            target_args.append("--force")
        byzerllm_main(target_args)

    elif args.command == "chat":
        byzerllm_main(
            [
                "query",
                "--model",
                instance_name,
                "--ray_address",
                args.ray_address,
                "--query",
                args.query,
            ]
        )



import logging
from typing import Any, Callable, Type, Union

import openai
from openai import (
    Completion,
    ChatCompletion,
    APITimeoutError,
    APIConnectionError,
    RateLimitError,
    APIError,
)

from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

logger = logging.getLogger(__name__)

CompletionClientType = Union[Type[Completion], Type[ChatCompletion]]


def _create_retry_decorator(max_retries: int) -> Callable[[Any], Any]:
    min_seconds = 4
    max_seconds = 10
    # Wait 2^x * 1 second between each retry starting with
    # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
    return retry(
        reraise=True,
        stop=stop_after_attempt(max_retries),
        wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
        retry=(
                retry_if_exception_type(APITimeoutError)
                | retry_if_exception_type(APIError)
                | retry_if_exception_type(APIConnectionError)
                | retry_if_exception_type(RateLimitError)
        ),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )


def completion_with_retry(is_chat_model: bool, max_retries: int, **kwargs: Any) -> Any:
    """Use tenacity to retry the completion call."""
    retry_decorator = _create_retry_decorator(max_retries=max_retries)

    @retry_decorator
    def _completion_with_retry(**kwargs: Any) -> Any:
        client = get_completion_endpoint(is_chat_model)
        return client.create(**kwargs)

    return _completion_with_retry(**kwargs)


async def async_completion_with_retry(is_chat_model: bool, max_retries: int, **kwargs: Any) -> Any:
    """Use tenacity to retry the async completion call."""
    retry_decorator = _create_retry_decorator(max_retries=max_retries)

    @retry_decorator
    async def _completion_with_retry(**kwargs: Any) -> Any:
        # Use OpenAI's async api https://github.com/openai/openai-python#async-api
        client = get_completion_endpoint(is_chat_model)
        return await client.acreate(**kwargs)

    return await _completion_with_retry(**kwargs)


def get_completion_endpoint(is_chat_model: bool) -> CompletionClientType:
    if is_chat_model:
        return openai.ChatCompletion
    else:
        return openai.Completion




import inspect
from dataclasses import dataclass
from typing import Callable, ClassVar, Dict, List, Type, TypeVar


handlers: Dict[str, List[Callable]] = {}

T = TypeVar("T", bound="BaseEvent")

class BaseEvent:
    __event_name__: ClassVar[str] = ""

    @classmethod
    def register(cls, handler):
        if cls not in handlers:
            handlers[cls] = []
        handlers[cls].append(handler)

    @classmethod
    def call_event(cls: Type[T], *args, **kwargs) -> T:
        if len(args) == 1 and type(args[0]) == cls:
            event = args[0]
        else:
            event = cls(*args, **kwargs)
        if event is None:
            event = cls()
        if not isinstance(event, BaseEvent):
            raise TypeError(
                "Expected event to be an instance of BaseEvent, got {}".format(
                    type(event)
                )
            )

        if cls not in handlers:
            return event

        for handler in handlers[cls]:
            handler(event)

        return event

    def __call__(self):
        fields = self.__dataclass_fields__
        results = []
        for field in fields:
            results.append(getattr(self, field))
        return results


@dataclass
class CancellableEvent(BaseEvent):
    cancelled = False


@dataclass
class SkippableEvent(BaseEvent):
    skip = False


def event_handler():
    def decorator(func: Callable[[BaseEvent], None]):
        sig = inspect.signature(func)
        args_types = sig.parameters.values()
        if len(args_types) < 1:
            return
        e, *_ = args_types
        e = e.annotation
        if issubclass(e, BaseEvent):
            e.register(func)

    return decorator



from typing import Dict,Any
from dataclasses import asdict
from byzerllm.utils.client import ByzerLLM
import json
import ray
import byzerllm
import os
from loguru import logger
from typing import Optional
from os.path import expanduser
from byzerllm import ByzerLLM,ByzerRetrieval
from dataclasses import dataclass
from pyjava.api.mlsql import PythonContext,RayContext
from byzerllm.apps.byzer_storage import env as env_detect
from byzerllm.utils.connect_ray import _check_java_version

@dataclass
class Env:
    llm:ByzerLLM
    retrieval:Optional[ByzerRetrieval]
    ray_context:RayContext           
     

def prepare_env(globals_info:Dict[str,Any],context:PythonContext)->Env:
    
    conf:Dict[str,str] = context.conf

    ray_address = conf.get("rayAddress","auto")
    enable_storage = conf.get("enable_storage","false") in ["true","True"]
    base_dir = conf.get("base_dir",None)
    storage_version = conf.get("storage_version",None)
    java_home = conf.get("java_home",None)

    version = storage_version or "0.1.11"        
    home = expanduser("~")        
    base_dir = base_dir or os.path.join(home, ".auto-coder")    
    code_search_path = None 
    storage_enabled = False
    
    if enable_storage:    
        
        if not os.path.exists(base_dir):
            logger.info(f"Byzer Storage not found in {base_dir}")
            base_dir = os.path.join(home, ".byzerllm")

        if os.path.exists(base_dir):            
            storage_enabled = True
        else:
            logger.info(f"Byzer Storage not found in {base_dir}")
               
        if storage_enabled:
            logger.info(f"Byzer Storage found in {base_dir}")            
            libs_dir = os.path.join(base_dir, "storage", "libs", f"byzer-retrieval-lib-{version}")        
            code_search_path = [libs_dir]
            logger.info(f"Connect and start Byzer Retrieval version {version}")     
        
    job_config = None
    env_vars = {}
    
    java_home=java_home if java_home else os.environ.get("JAVA_HOME")
    
    v = env_detect.detect_env() 
    if v.java_home and v.java_version == "21": 
        logger.info(f"JDK 21 will be used ({v.java_home})...")    
        java_home = v.java_home        
    
    if java_home:            
        path = os.environ.get("PATH")    
        env_vars = {"JAVA_HOME": java_home,
                    "PATH":f'''{os.path.join(java_home,"bin")}:{path}'''}        
        if code_search_path:
            if java_home:
                _check_java_version(java_home)
            job_config = ray.job_config.JobConfig(code_search_path=code_search_path,
                                                            runtime_env={"env_vars": env_vars})
    if not java_home and code_search_path:
       logger.warning("code_search_path is ignored because JAVA_HOME is not set") 
            
    ray_context = RayContext.connect(globals_info,ray_address,job_config=job_config)
        
    retrieval = None
    if storage_enabled:
        retrieval = ByzerRetrieval()
        retrieval.launch_gateway()
        
    llm = byzerllm.ByzerLLM()    
    return Env(llm=llm,
            retrieval=retrieval,
            ray_context=ray_context)

def chat(ray_context):   
    conf = ray_context.conf()
    udf_name = conf["UDF_CLIENT"] 
    
    input_value = [json.loads(row["value"]) for row in ray_context.python_context.fetch_once_as_rows()]
    
    llm = ByzerLLM()
    llm.setup_template(model=udf_name,template="auto")
    llm.setup_default_emb_model_name("emb")
    llm.setup_default_model_name(udf_name)
    llm.setup_extra_generation_params(udf_name,extra_generation_params={
        "temperature":0.01,
        "top_p":0.99
    })
    
    result = []
    for value in input_value:
        v = value.get("query",value.get("instruction",""))
        history = json.loads(value.get("history","[]"))
        
        for key in ["query","instruction","history","user_role","system_msg","assistant_role"]:
            value.pop(key,"")            
                
        conversations = history + [{
            "role":"user",
            "content":v
        }]
        t = llm.chat_oai(conversations=conversations,llm_config={**value})

        response = asdict(t[0])
        
        new_history =  history + [{
            "role":"user",
            "content":v
        }] + [{
            "role":"assistant",
            "content":response["output"]
        }]  

        response["history"] = new_history    
        
        result.append({"value":[json.dumps(response,ensure_ascii=False)]})
    
    ray_context.build_result(result) 


def deploy(infer_params:str,conf:Dict[str,str]):
    '''
    !byzerllm setup single;
    !byzerllm setup "num_gpus=4";
    !byzerllm setup "resources.master=0.001";
    run command as LLM.`` where 
    action="infer"
    and pretrainedModelType="llama"
    and localModelDir="/home/byzerllm/models/openbuddy-llama-13b-v5-fp16"
    and reconnect="true"
    and udfName="llama_13b_chat"
    and modelTable="command";
    '''
    infer_params = json.loads(infer_params)
    llm = ByzerLLM()
    
    num_gpus = float(conf.get("gpus_per_worker",conf.get("gpusPerWorker",0)))
    num_cpus = float(conf.get("cpus_per_worker",conf.get("cpusPerWorker",0.001)))
    num_workers = int(conf.get("num_workers",conf.get("numWorkers",1)))
    worker_concurrency = int(conf.get("worker_concurrency",conf.get("workerConcurrency",1)))

    pretrained_model_type = infer_params.get("pretrainedModelType","custom/auto")
    model_path = infer_params.get("localModelDir","")
    
    infer_params.pop("pretrainedModelType","")
    infer_params.pop("localModelDir","")
    infer_params.pop("udfName","")
    infer_params.pop("modelTable","")
    infer_params.pop("reconnect","")

    chat_name = conf["UDF_CLIENT"]
    
    llm.setup_num_workers(num_workers).setup_worker_concurrency(worker_concurrency)
    llm.setup_cpus_per_worker(num_cpus).setup_gpus_per_worker(num_gpus)

    llm.deploy(model_path=model_path,
            pretrained_model_type=pretrained_model_type,
            udf_name=chat_name,
            infer_params={
               **infer_params
            })            





# Adapted from
# https://github.com/skypilot-org/skypilot/blob/86dc0f6283a335e4aa37b3c10716f90999f48ab6/sky/sky_logging.py
# Adapted vLLM
"""Logging configuration for ByzerLLM."""

import logging
import os
import sys

BYZERLLM_CONFIGURE_LOGGING = int(os.getenv("BYZERLLM_CONFIGURE_LOGGING", "1"))

_FORMAT = "[%(levelname)s][%(asctime)s][%(filename)s:%(lineno)d] - %(message)s"
_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class NewLineFormatter(logging.Formatter):
    """Adds logging prefix to newlines to align multi-line messages."""

    def __init__(self, fmt, datefmt=None):
        logging.Formatter.__init__(self, fmt, datefmt)

    def format(self, record):
        msg = logging.Formatter.format(self, record)
        if record.message != "":
            parts = msg.split(record.message)
            msg = msg.replace("\n", "\r\n" + parts[0])
        return msg


_root_logger = logging.getLogger("byzerllm")
_default_handler = None


def _setup_logger():
    _root_logger.setLevel(logging.DEBUG)
    global _default_handler
    if _default_handler is None:
        _default_handler = logging.StreamHandler(sys.stdout)
        _default_handler.flush = sys.stdout.flush  # type: ignore
        _default_handler.setLevel(logging.INFO)
        _root_logger.addHandler(_default_handler)
    fmt = NewLineFormatter(_FORMAT, datefmt=_DATE_FORMAT)
    _default_handler.setFormatter(fmt)
    # Setting this will avoid the message
    # being propagated to the parent logger.
    _root_logger.propagate = False


# The logger is initialized when the module is imported.
# This is thread-safe as the module is only imported once,
# guaranteed by the Python GIL.
if BYZERLLM_CONFIGURE_LOGGING:
    _setup_logger()


def init_logger(name: str):
    # Use the same settings as above for root logger
    logger = logging.getLogger(name)
    logger.setLevel(os.getenv("LOG_LEVEL", "DEBUG"))
    if BYZERLLM_CONFIGURE_LOGGING:
        logger.addHandler(_default_handler)
        logger.propagate = False
    return logger



from dataclasses import dataclass, field
from typing import Optional


@dataclass
class CustomizedArguments:
    """
    一些自定义参数
    """
    max_seq_length: int = field(metadata={"help": "输入最大长度"})
    train_file: str = field(metadata={"help": "训练集"})
    model_name_or_path: str = field(metadata={"help": "预训练权重路径"})
    eval_file: Optional[str] = field(default="", metadata={"help": "the file of training data"})


@dataclass
class QLoRAArguments:
    """
    一些自定义参数
    """
    max_seq_length: int = field(metadata={"help": "输入最大长度"})
    train_file: str = field(metadata={"help": "训练集"})
    model_name_or_path: str = field(metadata={"help": "预训练权重路径"})
    task_type: str = field(default="", metadata={"help": "预训练任务：[sft, pretrain]"})
    eval_file: Optional[str] = field(default="", metadata={"help": "the file of training data"})
    lora_rank: Optional[int] = field(default=64, metadata={"help": "lora rank"})
    lora_alpha: Optional[int] = field(default=16, metadata={"help": "lora alpha"})
    lora_dropout: Optional[float] = field(default=0.05, metadata={"help": "lora dropout"})    

